####
# bomail.cli.send
#
# Send drafts.
####

import os
import sys
import mimetypes
import subprocess, email
import smtplib

from email import encoders
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart
from email.mime.audio import MIMEAudio
from email.mime.base import MIMEBase
from email.mime.image import MIMEImage

import bomail.config.config as config
import bomail.cli.mailfile as mailfile
import bomail.util.addr as addr

usage_str = """
Reads filename(s) from stdin one per line and sends them.
send.py -h to print this help.
"""


# adapted from python library example: https://docs.python.org/3/library/email-examples.html
def add_attachments(msg, attach_list):
  for filename in attach_list:
    if not os.path.exists(filename):
      raise Exception("Did not send. Could not attach (does not exist): " + filename)
    if not os.path.isfile(filename):
      raise Exception("Did not send. Could not attach (is not a file): " + filename)
    ctype, encoding = mimetypes.guess_type(filename)
    if ctype is None or encoding is not None:
      ctype = 'application/octet-stream'
    maintype, subtype = ctype.split('/', 1)
    if maintype == 'text':
      with open(filename) as fp:
        sub_msg = MIMEText(fp.read(), _subtype=subtype)
    elif maintype == 'image':
      with open(filename, 'rb') as fp:
        sub_msg = MIMEImage(fp.read(), _subtype=subtype)
    elif maintype == 'audio':
      with open(filename, 'rb') as fp:
        sub_msg = MIMEAudio(fp.read(), _subtype=subtype)
    else:
      with open(filename, 'rb') as fp:
        sub_msg = MIMEBase(maintype, subtype)
        sub_msg.set_payload(fp.read())
      encoders.encode_base64(sub_msg)
    sub_msg.add_header('Content-Disposition', 'attachment', filename=os.path.split(filename)[1])
    msg.attach(sub_msg)


def file_to_msg(filename, mgr):
  data = mgr.get_all(filename)
  attach_list = mgr.get_attachlist(filename)
  if len(attach_list) == 0:  # no attachments
    msg = MIMEText(data[mailfile.BODY_L], 'plain')
  else:
    msg = MIMEMultipart()
    msg.attach(MIMEText(data[mailfile.BODY_L], 'plain'))
    add_attachments(msg, attach_list)
  msg['Subject'] = data[mailfile.SUBJ_L]
  msg['From'] = data[mailfile.FROM_L]
  msg['To'] = data[mailfile.TO_L]
  msg['Date'] = email.utils.formatdate()
  msg['Message-ID'] = data[mailfile.MSG_ID_L]
  if len(data[mailfile.CC_L]) > 0:
    msg['CC'] = data[mailfile.CC_L]
  refs = data[mailfile.REFS_L]
  if len(refs) > 0:
    msg['In-Reply-To'] = refs.split(", ")[-1]
    msg['References'] = refs

  recip_lists = data[mailfile.TO_L].split(", ")
  if len(data[mailfile.CC_L]) > 0:
    recip_lists += data[mailfile.CC_L].split(", ")
  if len(data[mailfile.BCC_L]) > 0:
    recip_lists += data[mailfile.BCC_L].split(", ")
  recip_addrs = [addr.str_to_pair(a)[1] for a in recip_lists]
  return recip_addrs, msg


def connect_and_send(email_addr, filelist, mgr):
  try:
    user_ind = config.email_addrs.index(email_addr)
  except:
    return 0, "Error sending from " + email_addr + " (not one of my addresses)"

  try:
    if config.smtp_servernames[user_ind] == "localhost":
      serv = smtplib.SMTP("localhost")
    else:
      serv = smtplib.SMTP(config.smtp_servernames[user_ind], config.smtp_ports[user_ind])
      serv.ehlo()
      serv.starttls()
      serv.ehlo()
      serv.login(config.smtp_usernames[user_ind], config.smtp_passwords[user_ind])
  except Exception as e:
    with open(config.error_log_file, "a") as errf:
      errf.write("-----------------\n" + repr(e) + "\n")
    return 0, "Error connecting to server: " + str(e)

  for i, filename in enumerate(filelist):
    recip_addrs, msg = file_to_msg(filename, mgr)
    try:
      serv.sendmail(email_addr, recip_addrs, msg.as_string())
      if config.smtp_servernames[user_ind] == "localhost": # save the sent mail
        newfilename = mailfile.get_mailfilepath(mgr.get_all(filename))[1]
        newfilepath = os.path.join(config.save_sent_dir, newfilename)
        with open(newfilepath, "w") as savef:
          savef.write(msg.as_string())
    except Exception as e:
      serv.quit()
      with open(config.error_log_file, "a") as errf:
        errf.write("-----------------\n" + repr(e) + "\n")
      return i, "Error sending message #" + str(i) + " [" + filename + "]: " + str(e)

  serv.quit()
  return len(filelist), ""

 

# return num_successes, err_msg
def main(filelist, mgr):
  # sort them into different "from" addresses and connect to each server once
  fromaddr_to_fnames = {}
  for fname in filelist:
    fromline = mgr.get(fname, mailfile.FROM_L)
    frompr = addr.str_to_pair(fromline)
    fromaddr = frompr[1]
    if fromaddr not in fromaddr_to_fnames:
      fromaddr_to_fnames[fromaddr] = [fname]
    else:
      from_addr_to_fnames[fromaddr].append(fname)

  total_sent = 0
  for fromaddr,fnames in fromaddr_to_fnames.items():
    num, msg = connect_and_send(fromaddr, fnames, mgr)
    total_sent += num
    if len(msg) > 0:
      return total_sent, msg
  return total_sent, ""



def main_cli():
  if len(sys.argv) >= 2:
    print(usage_str)
    exit(0)

  filelist = [f.strip() for f in sys.stdin.readlines()]
  res, err = main(filelist, mailfile.MailMgr())
  print("Sent " + str(res) + "/" + str(len(filelist)) + " messages.")
  if res < len(filelist):
    print(err)


if __name__ == "__main__":
  main_cli()





####
# bomail.guistuff.get_text
#
# Getting input text for specific tasks.
####

import sys
import os
import time
import curses
import curses.textpad
import time
import textwrap, time

import bomail.config.config as config
import bomail.config.guiconfig as guicfg
import bomail.cli.mailfile as mailfile
import bomail.util.addr as addr
import bomail.util.search_opts as search_opts
import bomail.util.tags as tags
import bomail.guistuff.display as display

LIST_INSTR_LINES = [
    "Type or use up/down arrows to navigate.",
    "Tab: accept current line and go to next entry.",
    "Enter: accept current line and return list.",
    "Escape: cancel current line and return list.",
    "Ctrl-x: cancel entire process."]


#---------------------------------------------------------------------
# mini-screens to get input for specific tasks

def get_subj_line(gui, init_str):
  display.draw_bottom_text(gui, ["Enter subject line (ESC=cancel)."])
  height,width = gui.note_area.getmaxyx()
  return get_txt_line(gui.note_area, 0, 0, width, "", gui.attr)

def get_search_in_tab_txt(gui):
  display.draw_bottom_text(gui, ["Enter search query (ESC=cancel)."])
  height,width = gui.note_area.getmaxyx()
  return get_txt_line(gui.note_area, 0, 0, width, "", gui.attr)



sched_instructions = [
      "format: yyyy-mm-dd-HH:MM   (or any prefix)",
      "        p10d               (plus 10 days from now at 00:00)",
      "        p3.5w              (plus exactly 3.5 weeks from now)",
      "ESC=cancel.  units are y (year), w (week), d (day), H (hour), M (minute)."]


def get_sched_txt(gui):
  display.draw_bottom_text(gui, sched_instructions)
  display.redraw_note(gui, "")
  height, width = gui.note_area.getmaxyx()
  return get_txt_line(gui.note_area, 0, 0, width, "", gui.attr)

def get_tab_search_txt(gui, init_str):
  height,width = gui.tab_area.getmaxyx()
  gui.tab_area.clear()
  gui.tab_area.addstr(0, 0, "Enter to accept, ESC to cancel", gui.attr)
  for i,s in enumerate(search_opts.options_str.split("\n")):
    if i+2 >= height:
      break
    gui.tab_area.addstr(i + 2, 0, s[:width], gui.attr)
  gui.tab_area.refresh()
  height, width = gui.note_area.getmaxyx()
  display.draw_bottom_text(gui, "")
  return get_txt_line(gui.note_area, 0, 0, width, init_str, gui.attr)


def get_decoded(key):
  try:
    res = curses.unctrl(key).decode("utf-8")
  except:
    res = key
  return res


# return None if ESC is pressed
def get_txt_line(window, y, x, width, initstr, attr):
  s = initstr
  blankstr = " "*(width - x - 1)
  window.keypad(True)
  x_ind = len(s)
  while True:
    window.addstr(y, x, blankstr)
    window.addstr(y, x, s[:width-1], attr)
    window.chgat(y, x+x_ind, 1, curses.A_STANDOUT)  # cursor
    window.refresh()
    
    # read input and react
    key = window.getkey()
    code = get_decoded(key)
    if key == "KEY_ESC" or code == "^[":
      return None
    elif key in ["KEY_ENTER", "\r", "\n", "\n\r"]:
      return s
    elif key in ["KEY_TAB", "\t"]:
      x_ind = len(s)
    elif key == "KEY_BACKSPACE" or code == "^?":
      if x_ind > 0:
        s = s[:x_ind-1] + ("" if x_ind >= len(s) else s[x_ind:])
        x_ind -= 1
    elif key == "KEY_LEFT":
      if x_ind > 0:
        x_ind -= 1
    elif key == "KEY_RIGHT":
      if x_ind < len(s):
        x_ind += 1
    else:
      s = s[:x_ind] + key + ("" if x_ind >= len(s) else s[x_ind:])
      x_ind += 1
  

# return window for editing, window for results, height, width (of editing window)
def draw_list_top_and_instr(window, instr_lines):
  height, width = window.getmaxyx()
  toplines = len(instr_lines) + 3
  height -= toplines
  window.clear()
  for i,line in enumerate(instr_lines):
    window.addstr(i, 0, line[:width-1])
  window.hline(toplines-3, 0, "-", width)
  window.hline(toplines-1, 0, "-", width)
  window.refresh()
  new_window = curses.newwin(height, width, toplines + display.TOPLINES, 0)
  new_window.bkgd(" ", curses.color_pair(guicfg.DEFAULT_CLR_PAIR))
  new_window.keypad(1)

  results_win = curses.newwin(1, width, toplines + display.TOPLINES - 2, 0)
  results_win.bkgd(" ", curses.color_pair(guicfg.DEFAULT_CLR_PAIR))
  return new_window, results_win, height, width


# get_bold_location(elem_str, curr_str) gets the index in elem_str to start bolding
def draw_list(curr_index, height, width, results_win, window, curr_str, curr_list, result_str, elem_tostr, get_bold_location):
  offset = 0
  if curr_index >= height:
    offset = curr_index - height + 1
  results_win.clear()
  results_win.addstr(0, 0, result_str)
  results_win.refresh()
  window.clear()
  if len(curr_list) > 0:
    for j in range(min(len(curr_list)-offset, height)):
      elem_str = elem_tostr(curr_list[j + offset])
      window.addstr(j, 0, elem_str[:width-1])
      bld_start = get_bold_location(elem_str, curr_str)
      if bld_start < width:
        window.chgat(j, bld_start, min(len(curr_str), width-bld_start-1), curses.A_BOLD)
    window.chgat(curr_index-offset, 0, width, curses.A_STANDOUT)
    elem_str = elem_tostr(curr_list[curr_index])
    bld_start = get_bold_location(elem_str, curr_str)
    if bld_start < width:
      window.chgat(curr_index-offset, bld_start, min(len(curr_str), width-bld_start-1), curses.A_STANDOUT | curses.A_BOLD)
  else:
    # element not in the list
    window.addstr(0, 0, curr_str[:width-1], curses.A_BOLD)
  window.refresh()


def get_bold_start(addrstr, substr):
  if addrstr.startswith(substr):
    return 0
  else:
    return addrstr.index("<")


# return a list of recipient pairs [name, addr]
# Choose from the list of recip_pairs, or the whole address book if it is None.
# Do not allow choosing from already_recip_pairs.
def get_recips(window, addr_book, recip_pairs=None, already_recip_pairs=None, allow_new=True):
  # recips is a list of (count_pair, str, name_pair)
  # where str is the display version of name and email_addr
  # and count_pair is negated (so that sorting ascending gives highest count first)
  if recip_pairs is None:
    if already_recip_pairs is None:
      recips = [((-cnt[0], -cnt[1]), addr.pair_to_str(*pr), pr) for pr,cnt in addr_book.pair_to_count.items()]
    else:
      recips = [((-cnt[0], -cnt[1]), addr.pair_to_str(*pr), pr) for pr,cnt in addr_book.pair_to_count.items() if pr not in already_recip_pairs]
    # TODO: aliases!

  else: # given a list of recip_pairs
    if already_recip_pairs is None:
      recips = [((0, 0), addr.pair_to_str(*r), r) for r in recip_pairs]
    else:                
      recips = [((0, 0), addr.pair_to_str(*r), r) for r in recip_pairs if r not in already_recip_pairs]

  if len(recips) == 0 and not allow_new:
    return []

  instr = [
    "New recipients must be either:",
    "    Person Possibly-Full Name <email@addr.domain>",
    "    email@addr.domain"]
  window, results_win, height, width = draw_list_top_and_instr(window, instr)

  result_pairs = []

  # sort by send count, then receive count, then alphabetically
  recips.sort()
  eligible_cips = recips    # will narrow to exclude those already-entered
  curr_cips = eligible_cips
  curr_str = ""
  curr_index = 0
  while True:
    results_str = ", ".join([addr.pair_to_str(*r) for r in result_pairs])
    draw_list(curr_index, height, width, results_win, window, curr_str, curr_cips, results_str, lambda cip: cip[1], get_bold_start)

    # read input and react
#    key = window.getkey()
    key = window.getkey() #curses.keyname(ch).decode("utf-8")
    code = get_decoded(key)
    if key in ["\n", "\t"]:
      if len(curr_cips) == 0:
        if allow_new and len(curr_str) > 0:
          result_pairs.append(addr.str_to_pair(curr_str))
      else:
        result_pairs.append(curr_cips[curr_index][2])
      if key == "\n":
        break
      # reset, get next recipient
      curr_str = ""
      eligible_cips = [r for r in recips if r[2] not in result_pairs]
      curr_cips = eligible_cips
      curr_index = 0
    elif key == "KEY_ESC" or code == "^[":
      window.nodelay(True)    # a hack needed to detect ESC
      newch = window.getch()
      window.nodelay(False)
      if newch == -1: # it was ESC
        break
    elif len(key) <= 1 and code == "^X":
      return None
    elif key == "KEY_BACKSPACE" or key == "KEY_DELETE" or code == "^?":
      if len(curr_str) > 0:
        curr_str = curr_str[:-1]
        if curr_index > 0:
          prev_entry = curr_cips[curr_index]
        # keep a recip if curr_str is a prefix of either the long str or the email addr
        curr_cips = [r for r in eligible_cips if r[1].startswith(curr_str) or r[2][1].startswith(curr_str)]
        if curr_index > 0:
          curr_index = curr_cips.index(prev_entry)
    elif key == "KEY_UP":#ch == curses.KEY_UP:
      if curr_index > 0:
        curr_index -= 1
    elif key == "KEY_DOWN":#ch == curses.KEY_DOWN:
      if curr_index < len(curr_cips) - 1:
        curr_index += 1
    else:
      try:
        new_str = curr_str + key
        if not allow_new and itertools.none(new_str in cip[1] for cip in eligible_cips):
          continue
        curr_str = new_str
        if len(curr_cips) > 0:
          prev_highlight = curr_cips[curr_index]
        curr_cips = [r for r in curr_cips if r[1].startswith(curr_str) or r[2][1].startswith(curr_str)]
        if len(curr_cips) > 0 and prev_highlight in curr_cips:
          curr_index = curr_cips.index(prev_highlight)
        else:
          curr_index = 0
      except ValueError:
        pass
   
  addr_book.check_for_new(result_pairs)
  return result_pairs




# given a path, get a list of pairs (file, is_directory)
# for everything in the same directory and starting with path
def get_curr_poss(s):
    try:
      curr_dir = os.path.dirname(s)
      if curr_dir[-1] != "/":
        curr_dir += "/"
      res = []
      if curr_dir == s:
        res.append((s, True))
      for f in os.listdir(curr_dir):
        full = curr_dir + f
        if full.startswith(s):
          is_dir = os.path.isdir(full)
          res.append((full + "/" if is_dir else full, is_dir))
      res.sort()
      return res
    except:
      return []


# Global variable
most_recent_path = "/"

def get_path(window, mail_mgr):
  curr_str = most_recent_path
  
  instr_lines = [
    "Type or use up/down arrows.",
    "TAB: auto-complete.",
    "ENTER: accept current line and continue.",
    "ESC: cancel current line and return.",
    "Ctrl-x: cancel entire process."]
  window, results_win, height, width = draw_list_top_and_instr(window, instr_lines)

  curr_poss = get_curr_poss(curr_str)
  if len(curr_poss) == 0:
    curr_str = "/"
    curr_poss = get_curr_poss(curr_str)

  results = []
  curr_index = 0
  while True:
    draw_list(curr_index, height, width, results_win, window, curr_str, [p[0] for p in curr_poss], ", ".join(results), lambda s: s, lambda a,b: 0)
    key = window.getkey()
    code = get_decoded(key)
    if key in ["\n", "\t"]:
      if (key == "\n" or (len(curr_poss) == 1 and curr_str == curr_poss[curr_index][0])) and not curr_poss[curr_index][1]:
        # "autocomplete"
        results.append(curr_poss[curr_index][0])
      elif key == "\t":
        curr_str = os.path.commonprefix([p[0] for p in curr_poss])
        curr_poss = get_curr_poss(curr_str)
        cursor_ind = 0
        display_ind = 0
        continue
        ## autocomplete up to the next slash or end of current item
        ## then continue editing loop
        #try:
        #  i = curr_list[curr_index].index("/", len(curr_str))
        #  curr_str = curr_list[curr_index][:i] + "/"
        #except ValueError:
        #  curr_str = curr_list[curr_index] + "/"
        #continue
      # we just added a new result, continue
      curr_poss = get_curr_poss(os.path.dirname(curr_str))
      curr_str = curr_poss[0][0]
      curr_index = 0
    elif key == "KEY_ESC" or code == "^[":
      window.nodelay(True)    # a hack needed to detect ESC
      newch = window.getch()
      window.nodelay(False)
      if newch == -1: # it was ESC
        break
    elif len(key) <= 1 and code == "^X":
      return None
    elif key == "KEY_BACKSPACE" or key == "KEY_DELETE" or code == "^?":
      if len(curr_str) > 0:
        curr_str = curr_str[:-1]
        if curr_index > 0:
          prev_highlight = curr_poss[curr_index][0]
        curr_poss = get_curr_poss(curr_str)
        if curr_index > 0:
          strlist = [p[0] for p in curr_poss]
          curr_index = strlist.index(prev_highlight)
    elif key == "KEY_UP":#ch == curses.KEY_UP:
      if curr_index > 0:
        curr_index -= 1
    elif key == "KEY_DOWN":#ch == curses.KEY_DOWN:
      if curr_index < len(curr_poss) - 1:
        curr_index += 1
    else:
      try:
        if not any([p[0].startswith(curr_str + key) for p in curr_poss]):
          continue
        curr_str += key
        prev_highlight = curr_poss[curr_index][0] if len(curr_poss) > 0 else ""
        curr_poss = get_curr_poss(curr_str)
        strlist = [p[0] for p in curr_poss]
        if len(curr_poss) > 0 and prev_highlight in strlist:
          curr_index = strlist.index(prev_highlight)
        else:
          curr_index = 0
      except ValueError:
        pass
  return results



# returns None if cancelled, else a list of elements
# each selected from mainlist
# (or newly entered, if allow_new)
def get_sublist_from_list(window, mainlist, instr_lines, allow_new=True, taboo_chars=[]):
  instr_lines = instr_lines + LIST_INSTR_LINES
  window, results_win, height, width = draw_list_top_and_instr(window, instr_lines)

  results = []
  curr_str = ""
  curr_index = 0
  eligible_list = mainlist
  curr_list = mainlist
  while True:
    draw_list(curr_index, height, width, results_win, window, curr_str, curr_list, ", ".join(results), lambda s: s, lambda a,b: 0)
    key = window.getkey()
    code = get_decoded(key)
    if key in taboo_chars:
      continue
    elif key in ["\n", "\t"]:
      if len(curr_list) == 0:
        if allow_new and len(curr_str) > 0:
          results.append(curr_str[:-1] if curr_str[-1] == "/" else curr_str)
      elif key == "\n" or (len(curr_list) == 1 and curr_str == curr_list[curr_index]):
        # "autocomplete"
        results.append(curr_list[curr_index])
      elif key == "\t":
        # autocomplete up to the next slash or end of current item
        # then continue editing loop
        try:
          i = curr_list[curr_index].index("/", len(curr_str))
          curr_str = curr_list[curr_index][:i] + "/"
        except ValueError:
          curr_str = curr_list[curr_index] + "/"
        continue
      # we just added a new result, continue
      curr_str = ""
      eligible_list = [t for t in mainlist if t not in results]
      curr_list = eligible_list
      curr_index = 0
    elif key == "KEY_ESC" or code == "^[":
      window.nodelay(True)    # a hack needed to detect ESC
      newch = window.getch()
      window.nodelay(False)
      if newch == -1: # it was ESC
        break
    elif len(key) <= 1 and code == "^X":
      return None
    elif key == "KEY_BACKSPACE" or key == "KEY_DELETE" or code == "^?":
      if len(curr_str) > 0:
        curr_str = curr_str[:-1]
        if curr_index > 0:
          prev_highlight = curr_list[curr_index]
        curr_list = [s for s in eligible_list if s.startswith(curr_str)]
        if curr_index > 0:
          curr_index = curr_list.index(prev_highlight)
    elif key == "KEY_UP":#ch == curses.KEY_UP:
      if curr_index > 0:
        curr_index -= 1
    elif key == "KEY_DOWN":#ch == curses.KEY_DOWN:
      if curr_index < len(curr_list) - 1:
        curr_index += 1
    else:
      try:
        if not allow_new and not any([e.startswith(curr_str + key) for e in eligible_list]):
          continue
        curr_str += key
        prev_highlight = curr_list[curr_index] if len(curr_list) > 0 else ""
        curr_list = [s for s in curr_list if s.startswith(curr_str)]
        if len(curr_list) > 0 and prev_highlight in curr_list:
          curr_index = curr_list.index(prev_highlight)
        else:
          curr_index = 0
      except ValueError:
        pass
  return results



#--------------------------------------------------------------
# Tags

# The typed letters "curr_str" match the tag
# if any subtag starts with curr_str
# (where subtags start at a / or beginning of string)
def tag_matches(s, curr_str):
  try:
    a = s.index(curr_str)
  except:
    return False
  if a != 0 and s[a-1] != "/":
    return False
  return True

# return the lowest "tag level" that matches curr_str
# return -1 if none
def get_matching_level(s, curr_str):
  cs = curr_str if curr_str[-1] != "/" else curr_str[:-1]
  levels = s.split("/")
  for j,level in enumerate(levels):
    if level.startswith(cs):
      return j
  return -1

def get_tagsort_keyfunc(curr_str):
  def f(s):
    level = get_matching_level(s, curr_str)
    if level < 0:
      return (10000000, s)
    else:
      return (level, s)
  return f


def build_curr_list(eligible_list, curr_str):
  if len(curr_str) == 0:
    return eligible_list
  ans = [s for s in eligible_list if tag_matches(s, curr_str)]
  ans.sort(key = get_tagsort_keyfunc(curr_str))
  # add current string (extended with "/") if not already in list
  ext_curr = curr_str + ("/" if curr_str[-1] != "/" else "")
  if ext_curr not in ans:
    ans.append(ext_curr)
  return ans


def build_disp_list(curr_list, curr_str):
  # only display "first level", i.e. if we have dirname, then skip dirname/subname, etc.
  # UNLESS curr_str is that level (i.e. we are only getting subfolders of that level)
  disp_list = []
  for l in curr_list:
    # if this is a sub-heading of the previous item
    if len(disp_list) > 0 and l.startswith(disp_list[-1]):
      # if user has not entered all of previous item
      if len(curr_str) == 0 or not disp_list[-1].endswith(curr_str):
        continue  # don't display this one
    disp_list.append(l)
  return disp_list


# returns None if cancelled, else a list of tags
# selected from taglist
# (or newly entered, if allow_new)
# This is a customized version of get_sublist_from_list
def get_tags(window, taglist, allow_new=True):
  instr_lines = [
    "Type to narrow list or create new; navigate with up/down arrows.",
    "Tab: auto-complete.",
    "Space: accept current line and move to next item.",
    "Enter: accept current line and return list.",
    "Escape: cancel current line and return list.",
    "Ctrl-x: cancel entire process."]

  mainlist = [t + "/" for t in taglist]
  taboo_chars = [","]
  window, results_win, height, width = draw_list_top_and_instr(window, instr_lines)

  results = []
  curr_str = ""
  curr_index = 0
  eligible_list = mainlist
  curr_list = mainlist
  disp_list = build_disp_list(curr_list, curr_str)
  while True:
    draw_list(curr_index, height, width, results_win, window, curr_str, disp_list, ", ".join(results), lambda s: s, lambda a,b: a.index(b))
    key = window.getkey()
    code = get_decoded(key)
    if key in taboo_chars:
      pass
    elif key in ["\n", " "]:
      # add new entry to results
      if len(disp_list) == 0:
        if allow_new and len(curr_str) > 0:
          results.append(curr_str[:-1] if curr_str[-1] == "/" else curr_str)
      else:
        # use current line, but without trailing "/"
        results.append(disp_list[curr_index][:-1])
      if key == "\n":
        break
      # we just added a new result, continue
      curr_str = ""
      eligible_list = [t for t in mainlist if t not in results]
      curr_list = eligible_list
      disp_list = build_disp_list(curr_list, curr_str)
      curr_index = 0
    elif key == "\t":
      # autocomplete up to the next slash or end of current item
      try:
        if len(disp_list) > 0:
          a = disp_list[curr_index].index(curr_str)
          b = disp_list[curr_index].index("/", a + len(curr_str))
          # works if b == -1
          curr_str = disp_list[curr_index][a:b] + "/"
      except:
        curr_str = disp_list[curr_index]
      prev_highlight = disp_list[curr_index] if len(disp_list) > 0 else ""
      curr_list = build_curr_list(eligible_list, curr_str)
      disp_list = build_disp_list(curr_list, curr_str)
      if len(disp_list) > 0 and prev_highlight in disp_list:
        curr_index = disp_list.index(prev_highlight)
      else:
        curr_index = 0
    elif key == "KEY_ESC" or code == "^[":
      window.nodelay(True)    # a hack needed to detect ESC
      newch = window.getch()
      window.nodelay(False)
      if newch == -1: # it was ESC
        break
    elif len(key) <= 1 and code == "^X":
      return None
    elif key == "KEY_BACKSPACE" or key == "KEY_DELETE" or code == "^?":
      if len(curr_str) > 0:
        curr_str = curr_str[:-1]
        if curr_index > 0:
          prev_highlight = disp_list[curr_index]
        curr_list = build_curr_list(eligible_list, curr_str)
        disp_list = build_disp_list(curr_list, curr_str)
        if curr_index > 0:
          curr_index = disp_list.index(prev_highlight)
    elif key == "KEY_UP":#ch == curses.KEY_UP:
      if curr_index > 0:
        curr_index -= 1
    elif key == "KEY_DOWN":#ch == curses.KEY_DOWN:
      if curr_index < len(disp_list) - 1:
        curr_index += 1
    else:
      try:
        if not allow_new and not any([tag_matches(e, curr_str + key) for e in eligible_list]):
          continue
        curr_str += key
        prev_highlight = disp_list[curr_index] if len(disp_list) > 0 else ""
        curr_list = build_curr_list(eligible_list, curr_str)
        disp_list = build_disp_list(curr_list, curr_str)
        if len(disp_list) > 0 and prev_highlight in disp_list:
          curr_index = disp_list.index(prev_highlight)
        else:
          curr_index = 0
      except ValueError:
        pass
  return results




####
# bomail.util.search_opts
#
# Parse search options!
####

import os
import shlex
import datetime

import bomail.cli.mailfile as mailfile
import bomail.util.tags as tags
import bomail.util.addr as addr
import bomail.util.datestuff as datestuff
from bomail.util.listmail import ListOpts

options_str = """
All arguments are optional:
 -h               [print this help]
 -n 200           only return first 200 results
 -sortold         list by oldest (default is by newest)

 -a datestr       after date. datestr can be e.g. yyyy-mm-dd
 -b datestr       before date

 -open            state is open
 -scheduled       state is scheduled
 -closed          state is closed
 -draft           is a draft
 -sent            is sent from me
 -attach          has an attachment

 -notags          has no tags
 -to-me           addressed to any of my aliases in config file

Use multiple times to match any, use "str1, str2" to match all:
 -t str           has str as a tag
 -nt str          does not have str as a tag
 -to str          str is in to, cc, or bcc field
 -from str        str is in from field
 -subject str     str is in subject field
 query            query is in email somewhere
"""


def with_quotes(q):
  return ('"' if q[0] != '"' else '') + q + ('"' if q[-1] != '"' else '')

# add query q to search string
def get_new_search_str(old_str, q):
  sq = SearchQuery()
  args = shlex.split(old_str)
  sq.parse(args)

  # if no search query in old string, just append this one
  if len(sq.querylist_list) == 0:
    return old_str + " " + with_quotes(q)

  # otherwise, add it to every search query
  ind = 0
  outargs = []
  while True:
    if ind >= len(args):
      break
    if args[ind] in ['-n', '-a', '-b', '-t', '-nt', '-to', '-from', '-subject']:
      outargs += args[ind:ind+2]
      ind += 2
    elif args[ind] in ['-sortold', '-open', '-scheduled', '-closed', '-draft', '-sent', '-attach', '-notags', '-to-me']:
      outargs += args[ind:ind+1]
      ind += 1
    else:
      outargs.append(with_quotes(args[ind] + ", " + q))
      ind += 1
  return " ".join(outargs)



class SearchQuery:
  def __init__(self, liststep=True):
    self.max_num = -1  # indiciates unlimited
    self.reverse = False

    self.listopts = ListOpts() if liststep else None
    self.orig_after_str = None
    self.orig_before_str = None
    self.after_str = None
    self.before_str = None

    # Can be found by listing (without opening files)
    self.open = None
    self.scheduled = None
    self.draft = None
    self.attach = None

    # Cannot be found by listing
    self.sent = None
    self.closed = None

    self.tagset_list = []
    self.not_tagset_list = []
    self.tolist_list = []
    self.fromlist_list = []
    self.subjectlist_list = []

    self.notags = None  # or True
    self.tome = None    # or True

    self.querylist_list = []


  def parse(self, old_arglist):
    arglist = list(old_arglist)
    if "-n" in arglist:
      i = arglist.index("-n")
      del arglist[i]
      if len(arglist) > i:
        try:
          self.max_num = int(arglist[i])
        except:
          self.max_num = 0
        del arglist[i]

    if "-sortold" in arglist:
      self.reverse = True
      arglist.remove("-sortold")

    if "-a" in arglist:
      i = arglist.index("-a")
      del arglist[i]
      if len(arglist) > i:
        s = arglist[i]
        self.orig_after_str = s
        del arglist[i]
    if "-b" in arglist:
      i = arglist.index("-b")
      del arglist[i]
      if len(arglist) > i:
        s = arglist[i]
        self.orig_before_str = s
        del arglist[i]
  
    # check draft first because it comes from a separate list
    # so it is always handled by listmail, not filter
    if "-draft" in arglist:
      if self.listopts is not None:
        self.listopts.source = "draft"
      else:
        self.draft = True
      arglist.remove("-draft")
    if "-open" in arglist:
      if self.listopts is not None and self.listopts.source == "all":
        self.listopts.source = "open"
      else:
        self.open = True
      arglist.remove("-open")
    if "-scheduled" in arglist:
      if self.listopts is not None and self.listopts.source == "all":
        self.listopts.source = "scheduled"
      else:
        self.scheduled = True
      arglist.remove("-scheduled")
    if "-attach" in arglist:
      if self.listopts is not None and self.listopts.source == "all":
        self.listopts.source = "attach"
      else:
        self.attach = True
      arglist.remove("-attach")

    if "-sent" in arglist:
      self.sent = True
      arglist.remove("-sent")
    if "-closed" in arglist:
      self.closed = True
      arglist.remove("-closed")

    while "-t" in arglist:
      i = arglist.index("-t")
      del arglist[i]
      if len(arglist) < i:
        break
      tagstr = arglist[i]
      self.tagset_list.append(tags.get_tagset_from_str(tagstr, include_folders=False))
      del arglist[i]

    while "-nt" in arglist:
      i = arglist.index("-nt")
      del arglist[i]
      if len(arglist) < i:
        break
      tagstr = arglist[i]
      self.not_tagset_list.append(tags.get_tagset_from_str(tagstr, include_folders=False))
      del arglist[i]

    while "-to" in arglist:
      i = arglist.index("-to")
      del arglist[i]
      if len(arglist) < i:
        break
      addrlist = arglist[i].split(",")
      self.tolist_list.append([s.strip() for s in addrlist])
      del arglist[i]

    while "-from" in arglist:
      i = arglist.index("-from")
      del arglist[i]
      if len(arglist) < i:
        break
      addrlist = arglist[i].split(",")
      self.fromlist_list.append([s.strip() for s in addrlist])
      del arglist[i]

    while "-subject" in arglist:
      i = arglist.index("-subject")
      del arglist[i]
      if len(arglist) < i:
        break
      strlist = arglist[i].split(",")
      self.subjectlist_list.append([s.strip() for s in strlist])
      del arglist[i]

    if "-notags" in arglist:
      self.notags = True
      arglist.remove("-notags")

    if "-to-me" in arglist:
      self.tome = True
      arglist.remove("-to-me")

    if len(arglist) > 0:
      self.querylist_list = [[s.strip() for s in quer.split(",")] for quer in arglist]



  def do_compile(self):
    slist = ["matched_inds = []\n"]
    slist.append("for ind,f in enumerate(filelist):\n")
    
    slist.append("  data = mgr.get_all(f)\n")

    # this is efficient so do it first
    if self.notags:
      slist.append("  if len(data[mailfile.TAGS_L]) > 0:\n")
      slist.append("    continue\n")

    need_date = self.after_str is not None or self.before_str is not None
    if need_date:
      slist.append("  datestr = data[mailfile.DATE_L]\n")
      slist.append("  if not datestr_matches(datestr, self.after_str, self.before_str):\n")
      slist.append("    continue\n")

    if self.open is not None:
      slist.append("  if not data[mailfile.STATE_L].startswith(\"open\"):\n")
      slist.append("    continue\n")
    if self.scheduled is not None:
      slist.append("  if not data[mailfile.STATE_L].startswith(\"scheduled\"):\n")
      slist.append("    continue\n")
    if self.closed is not None:
      slist.append("  if not data[mailfile.STATE_L].startswith(\"closed\"):\n")
      slist.append("    continue\n")
    if self.draft is not None:
      slist.append("  if f[-5:] != \"draft\":\n")
      slist.append("    continue\n")
    if self.sent is not None:
      slist.append("  if not data[mailfile.SENT_L] == \"True\":\n")
      slist.append("    continue\n")
    if self.attach is not None:
      slist.append("  if len(data[mailfile.ATTACH_L]) == 0:\n")
      slist.append("    continue\n")

    if self.tome is not None:
      slist.append("  prlist_list = [addr.str_to_pairlist(data[j]) for j in [mailfile.TO_L, mailfile.CC_L, mailfile.BCC_L]]\n")
      slist.append("  if not any([any([addr.is_pair_me(pr) for pr in prlist]) for prlist in prlist_list]):\n")
      slist.append("    continue\n")

    # need to match all of a set
    if len(self.tagset_list) > 0 or len(self.not_tagset_list) > 0:
      slist.append("  data_tagset = tags.get_tagset_from_str(data[mailfile.TAGS_L], include_folders=True)\n")
    if len(self.tagset_list) > 0:
      slist.append("  if not any([ts.issubset(data_tagset) for ts in self.tagset_list]):\n")
      slist.append("    continue\n")
    # if we match all of a set, then skip
    if len(self.not_tagset_list) > 0:
      slist.append("  if any([nts.issubset(data_tagset) for nts in self.not_tagset_list]):\n")
      slist.append("    continue\n")
    # need to match all of a set
    if len(self.tolist_list) > 0:
      slist.append("  if not any([all([s in data[mailfile.TO_L] or s in data[mailfile.CC_L] or s in data[mailfile.BCC_L] for s in tolist]) for tolist in self.tolist_list]):\n")
      slist.append("    continue\n")
    if len(self.fromlist_list) > 0:
      slist.append("  if not any([all([s in data[mailfile.FROM_L] for s in fromlist]) for fromlist in self.fromlist_list]):\n")
      slist.append("    continue\n")
    if len(self.subjectlist_list) > 0:
      slist.append("  if not any([all([s in data[mailfile.SUBJ_L] for s in subjectlist]) for subjectlist in self.subjectlist_list]):\n")
      slist.append("    continue\n")

    # skip if none of the queries have all their strings somewhere
    if len(self.querylist_list) > 0:
      slist.append("  if not any([all([any([q in s for s in data]) for q in qlist]) for qlist in self.querylist_list]):\n")
      slist.append("    continue\n")

    slist.append("  matched_inds.append(ind)\n")
    # stop if we've found enough!
    if self.max_num > 0:
      slist.append("  if len(matched_inds) >= " + str(self.max_num) + ":\n")
      slist.append("    break\n")

    with open("compile_log.txt", "w") as myf:
      myf.write("".join(slist))

    return compile("".join(slist), '<string>', 'exec')
    

  # convert relative dates and time zones to absolute UTC
  def interpret_dates(self):
    self.after_str = None
    self.before_str = None
    if self.listopts is not None:
      self.listopts.after_str = None
      self.listopts.before_str = None

    if self.orig_after_str is not None:
      self.after_str = datestuff.parse_schedstr_to_utcstr(self.orig_after_str)
      if self.listopts is not None:
        self.listopts.after_str = self.after_str[:10]
        if self.orig_after_str[0] in ["p","m"]:
          if not any([s in self.orig_after_str for s in [".","H","M"]]):
            self.after_str = None  # listmail can take care of it
        else:
          if len(self.orig_after_str) <= 10:
            self.after_str = None
 
    if self.orig_before_str is not None:
      self.before_str = datestuff.parse_schedstr_to_utcstr(self.orig_before_str)
      if self.listopts is not None:
        self.listopts.before_str = self.before_str[:10]
        if self.orig_before_str[0] in ["p","m"]:
          if not any([s in self.orig_before_str for s in [".","H","M"]]):
            self.before_str = None  # listmail can take care of it
        else:
          if len(self.orig_before_str) <= 10:
            self.before_str = None
                

  # list all mail matching basic part of query
  def listmail(self, mgr):
    self.interpret_dates()
    filelist = self.listopts.listmail(mgr)
    if self.reverse:
      return reversed(filelist)
    return filelist

  # return list of indices matching our query on filelist
  # which is assumed to have passed listmail testing first!
  def filter(self, mgr, filelist):
    self.interpret_dates()
    # check if we need to do anything!
    if all([x is None for x in [self.after_str, self.before_str, self.open, self.scheduled, self.closed, self.draft, self.sent, self.attach, self.notags, self.tome]]):
      if all([len(x) == 0 for x in [self.tagset_list, self.not_tagset_list, self.tolist_list, self.fromlist_list, self.subjectlist_list, self.querylist_list]]):
        return list(range(len(filelist)))

    compiled_obj = self.do_compile()
    namespace ={"self": self, "addr": addr, "datestr_matches": datestuff.datestr_matches, "mailfile": mailfile, "mgr": mgr, "tags": tags, "filelist": filelist}
    exec(compiled_obj, namespace)
    return namespace["matched_inds"]


  def search(self, mgr):
    filelist = self.listmail(mgr)
    matched_inds = self.filter(mgr, filelist)
    result = [filelist[i] for i in matched_inds] 
    if self.max_num > 0:
      return result[:self.max_num]
    return result





####
# bomail.cli.compose
#
# Compose new emails as blank, reply, forward.
####

import sys
import string
import traceback
import os, random
import dateutil, dateutil.parser
import email, email.utils, uuid

import bomail.config.config as config
import bomail.cli.mailfile as mailfile
import bomail.cli.chstate as chstate
import bomail.util.addr as addr
import bomail.util.util as util
import bomail.util.datestuff as datestuff

usage_str = """Use with option -h to print this help.
There are two forms of usage.

1. To compose a new email from scratch, use no options or any combination of:
  -to "example here"              add "example string" to recips (may be comma-sep list)
  -subj "example here"            set subject line to "example here"
  -cc "example here"              add "example here" to cc (may be comma-sep list)
  -bcc "example here"             ditto for bcc

2. To compose based on an existing email, use one of:
  -ro yyyy/mm-dd/msgid            reply-one to msgid
  -ra yyyy/mm-dd/msgid            reply-all to msgid
  -rr yyyy/mm-dd/msgid            reply-recipients to msgid
  -f yyyy/mm-dd/msgid "example"   forward msgid to "example"

(Both 1 and 2.) Optional, add:
  -body "body of email here"        
"""


def create(data, mgr):
  draft_dir, ext = mailfile.get_draftfilepath(data)
  filename = draft_dir + ext
  os.makedirs(draft_dir, exist_ok=True)
  chstate.create([filename], [data], mgr)
  chstate.make_open([filename], mgr)
  mgr.ensure_loaded(filename)
  return filename


def get_blank_draft_data():
  new_data = [""]*len(mailfile.fields)
  new_data[mailfile.FROM_L] = addr.pair_to_str(config.names[0], config.email_addrs[0])
  new_data[mailfile.DATE_L] = datestuff.get_local_nowstr()
  new_data[mailfile.SENT_L] = "True"
  new_data[mailfile.STATE_L] = "open"
  # create a message id, at least for now
  new_data[mailfile.MSG_ID_L] = email.utils.make_msgid(uuid.uuid4().hex)
  return new_data


# for a reply or forward, build new data from the old one
# if quote, include previous email message as quoted text
def get_old_and_new_data(msg_file, quote):
  old_data = mailfile.file_to_data(msg_file)
  new_data = get_blank_draft_data()
  # reply from the address it is addressed to, if possible
  for name,email_addr in reversed(list(zip(config.names,config.email_addrs))):
    if any([email_addr in old_data[j] for j in [mailfile.TO_L, mailfile.CC_L,mailfile.BCC_L]]):
      new_data[mailfile.FROM_L] = addr.pair_to_str(name, email_addr)
  new_data[mailfile.REFS_L] = old_data[mailfile.MSG_ID_L] + ", " + old_data[mailfile.REFS_L]
  new_data[mailfile.TAGS_L] = old_data[mailfile.TAGS_L] if config.autotag_draft_replies else ""
  prev_date_obj = dateutil.parser.parse(old_data[mailfile.DATE_L])
  prev_date_str = prev_date_obj.strftime(config.quote_date_fmt)
  prev_time_str = prev_date_obj.strftime(config.quote_time_fmt)
  if quote:
    body_lines = old_data[mailfile.BODY_L].split("\n")
    body_str = config.quote_line_prefix + ("\n" + config.quote_line_prefix).join(body_lines)
    new_data[mailfile.BODY_L] = config.quote_format.replace("%date", prev_date_str).replace("%time", prev_time_str).replace("%from", old_data[mailfile.FROM_L]).replace("%body", body_str)
  return old_data, new_data
  

def get_subject_with(old_sub, prefix):
  return old_sub if old_sub.startswith(prefix) else prefix + old_sub


def reply_one(msg_file, mgr, quote):
  old_data, new_data = get_old_and_new_data(msg_file, quote)
  new_data[mailfile.SUBJ_L] = get_subject_with(old_data[mailfile.SUBJ_L], "Re: ")
  new_data[mailfile.TO_L] = old_data[mailfile.FROM_L]
  return create(new_data, mgr)


def reply_all(msg_file, mgr, quote, include_sender=True):
  old_data, new_data = get_old_and_new_data(msg_file, quote)
  new_data[mailfile.SUBJ_L] = get_subject_with(old_data[mailfile.SUBJ_L], "Re: ")
  to_list = [s for s in old_data[mailfile.TO_L].split(", ") if not addr.is_str_me(s)]
  to_list += [s for s in old_data[mailfile.CC_L].split(", ") if not addr.is_str_me(s)]
  if include_sender:
    to_list = old_data[mailfile.FROM_L].split(", ") + to_list
  new_data[mailfile.TO_L] = ", ".join(to_list)
  return create(new_data, mgr)


def forward(msg_file, to_str, mgr, quote=True):
  old_data, new_data = get_old_and_new_data(msg_file, quote)
  new_data[mailfile.SUBJ_L] = get_subject_with(old_data[mailfile.SUBJ_L], "Fwd: ")
  new_data[mailfile.TO_L] = to_str
  return create(new_data, mgr)


def new_compose(subj_str, to_str, cc_str, bcc_str, bodystr, mgr):
  d = get_blank_draft_data()
  d[mailfile.TO_L] = to_str
  d[mailfile.CC_L] = cc_str
  d[mailfile.BCC_L] = bcc_str
  d[mailfile.SUBJ_L] = subj_str
  d[mailfile.BODY_L] = bodystr
  return create(d, mgr)


def main_cli():
  if "-h" in sys.argv:
    print(usage_str)
    exit(0)
  try:
    args = sys.argv
    bodystr = ""
    mgr = mailfile.MailMgr()
    if len(args) <= 1:
      print(new_compose("", "", "", "", mgr))
      sys.exit(0)

    if "-body" in args:
      ind = args.index("-body")
      if ind < len(args) - 1:
        bodystr = args[ind+1]
        args = args[:ind] + args[ind+2:]

    if args[1] == "-ro":
      print(reply_one(args[2], bodystr, mgr))
    elif args[1] == "-ra":
      print(reply_all(args[2], bodystr, mgr))
    elif args[1] == "-rr":
      print(reply_all(args[2], bodystr, mgr, include_sender=False))
    elif args[1] == "-f":
      print(forward(args[2], args[3], bodystr, mgr, quote=True))
    else:
      to_str_list = []
      cc_str_list = []
      bcc_str_list = []
      subj_str = ""
      i = 1
      while i < len(args):
        if args[i] == "-to":
          to_str_list.append(args[i+1])
          i += 1
        elif args[i] == "-subj":
          subj_str = args[i+1]
          i += 1
        elif args[i] == "-cc":
          cc_str_list.append(args[i+1])
          i += 1
        elif args[i] == "-bcc":
          bcc_str_list.append(args[i+1])
          i += 1
        i += 1
      print(new_compose(subj_str, ", ".join(to_str_list), ", ".join(cc_str), ", ".join(bcc_str), bodystr, mgr))
  except Exception:
    traceback.print_exc()
    print("")
    print(usage_str)


if __name__ == "__main__":
  main_cli()




####
# bomail.util.mime_to_mailfile
#
# Converts a file in MIME format to bomail's mailfile format.
####


import sys, email, uuid

import bomail.config.config as config
import bomail.cli.mailfile as mailfile
import bomail.util.addr as addr
import bomail.util.datestuff as datestuff

usage_str = """
Reads a MIME email from file and writes a mailfile to stdout.
Usage:
  mime_to_mailfile.py -h        # print this help
  mime_to_mailfile.py filename
"""


def clean(s, replace_newline_with = " "):
  if s is None:
    return ""
  result = ""
  for (h, charset) in email.header.decode_header(s):
      if isinstance(h, bytes):
        try:
          result += str(h, encoding = ("utf-8" if charset is None else charset), errors = "ignore")
        except:
          result += str(h)
      elif isinstance(h, str):
        result += h
  c = replace_newline_with
  return result.replace("\r\n", c).replace("\r", c).replace("\n", c)

def clean_addrs(unclean_addr_str):
  addr_str = clean(unclean_addr_str)
  prlist = addr.str_to_pairlist(addr_str)
  return addr.pairlist_to_str(prlist)

# m is a python email.message presumably of type "text/"
def get_plain_body(m):
  charset = m.get_content_charset()
  if charset is None:
    return str(m.get_payload(decode=True))
  return str(m.get_payload(decode=True), str(charset), "ignore")

def was_sent(m, prlist):
  return any([addr.is_pair_me(pr) for pr in prlist])


# m is a python email.message object
def get_body(m):
  if m.is_multipart():
    return get_multipart_body(m)
  else:
    return get_part_body(m)

def get_multipart_body(m):
  m_type = m.get_content_type()
  if m_type == "multipart/alternative":
    parts = m.get_payload()
    part_types = [p.get_content_type() for p in parts]
    if "text/plain" in part_types:
      return get_plain_body(parts[part_types.index("text/plain")])
    elif "text/html" in part_types:
      return get_plain_body(parts[part_types.index("text/html")])
    else:
      return ""
  else:
    # do the same thing for all of these:
    # multipart/mixed
    # multipart/related 
    # multipart/digest
    # multipart/report
    # message/rfc822
    # TODO: multipiart/encrypted
    # TODO: multipart/signed: check/verify signature if we have the public key
    # others?
    return "\n".join(get_body(p) for p in m.get_payload())

def get_part_body(m):
  m_type = m.get_content_type()
  if m_type == "text/plain":
    return get_plain_body(m)
  elif m_type == "text/html":
    return get_plain_body(m)
  else:
    return ""


def get_references(m, filename):
  m_id = get_msgid(m, filename)
  reply_to = ""
  if 'in-reply-to' in m:
    ref = m['in-reply-to']
    try:
      ref = ref[ref.index("<")+1:]
      right = ref.index(">")
      reply_to = ref[:right].strip()
      if reply_to == m_id:
        reply_to = ""
    except ValueError:
        pass

  reflist = []
  if 'references' in m:
    refs = m['references']
    while len(refs) > 2:
      try:
        refs = refs[refs.index("<")+1:]
        right = refs.index(">")
        new_ref = refs[:right].strip()
        if new_ref not in reflist and new_ref != reply_to and new_ref != m_id:
          reflist.append(new_ref)
        refs = refs[right+1:]
      except ValueError:
        break
  if len(reply_to) > 0:
    reflist.append(reply_to)
  return ", ".join("<" + r + ">" for r in reflist)


# from pymail object
def get_msgid(m, filename):
  attempt = clean(m['message-id']).strip()
  if len(attempt) == 0:
    attempt = "couldnotfindmessageidgenerated" + email.utils.make_msgid(uuid.uuid4().hex)
    with open(config.error_log_file, "a") as f:
      f.write("Could not parse message id for file: " + filename + "\n\n")
  return attempt

def main(filename):
  with open(filename) as f:
    m = email.message_from_file(f)
  return get_data(m, filename)

# m is a python email object
# return data array
def get_data(m, filename):
  data = ["" for l in mailfile.fields]
  from_prlist = addr.str_to_pairlist(clean(m['from']))

  data[mailfile.FROM_L] = addr.pairlist_to_str(from_prlist)
  data[mailfile.TO_L] = clean_addrs(m['to'])
  data[mailfile.CC_L] = clean_addrs(m['cc'])
  data[mailfile.SUBJ_L] = clean(m['subject'])
  data[mailfile.DATE_L] = datestuff.parse_email_date(clean(m['date'])).isoformat()
  data[mailfile.STATE_L] = "open"
  data[mailfile.MSG_ID_L] = get_msgid(m, filename)
  data[mailfile.REFS_L] = get_references(m, filename)
  data[mailfile.SENT_L] = str(was_sent(m, from_prlist))
  data[mailfile.BODY_L] = get_body(m)
  return data


if __name__ == "__main__":
  if "-h" in sys.argv or len(sys.argv) <= 1:
    print(usage_str)
    exit(0)
  data = main(sys.argv[1])
  s = mailfile.data_to_str(data)
  print(s)




####
# bomail.cli.search
#
# Command-line interface to search.
####

import sys
import os, subprocess, shlex
import textwrap

import bomail.config.config as config
import bomail.cli.mailfile as mailfile
import bomail.cli.chstate as chstate
import bomail.util.search_opts as search_opts

my_usage_str = """
Searches through mail for files matching given arguments.

"""
usage_str = my_usage_str + search_opts.options_str


def is_valid_searchstr(arg_str):
  try:
    shlex.split(arg_str)
    return True
  except:
    return False


# return list of filenames
def search_argstr(arg_str, mgr):
  # note: shlex.split is quite slow, avoid calling it many times per second
  return search_arglist(shlex.split(arg_str), mgr)

# return list of filenames
def search_arglist(args, mgr):
  query = search_opts.SearchQuery()
  query.parse(args)
  return query.search(mgr)

# return list of filenames
def filter_argstr(arg_str, mgr, filelist):
  return filter_arglist(shlex.split(arg_str), mgr, filelist)

# return list of filenames
def filter_arglist(args, mgr, filelist):
  query = search_opts.SearchQuery(liststep=False)
  query.parse(args)
  inds = query.filter(mgr, filelist)
  return [filelist[i] for i in inds]
  

def main_cli():
  args = sys.argv[1:]
  if "-h" in args:
    print(usage_str)
    exit()
  filelist = search_arglist(args, mailfile.MailMgr())
  print("\n".join(filelist))


if __name__ == "__main__":
  main_cli()




####
# bomail.util.merge_lines
#
# Merges sorted lines into a sorted file.
####

import sys, os, subprocess
import tempfile
from subprocess import PIPE, Popen

import bomail.util.util as util

usage_str = """
Reads a list of lines from stdin (one per line) and merges them into the file.
Assumes the file and input lines are sorted.

Use:
    merge_lines.py listfile         # sort in lexicographic order
    merge_lines.py -r listfile      # sort in reverse order
    merge_lines.py -h               # print this help
"""


def do(filename, lines, reverse=False):
  if reverse:
    comp = lambda a,b: a > b
  else:
    comp = lambda a,b: a <= b
  if not os.path.exists(filename):
    os.makedirs(os.path.dirname(filename), exist_ok=True)
    with open(filename, "a"):
      pass

  if len(lines) == 0:
    return
  
  with open(filename) as fromf, tempfile.NamedTemporaryFile(dir=".", mode="w", delete=False) as tof:
    lines_ind = 0
    next_fileline = fromf.readline()
    while True:
      if len(next_fileline) == 0 and lines_ind >= len(lines):
        break
      elif lines_ind >= len(lines) or (len(next_fileline) > 0 and comp(next_fileline, lines[lines_ind])):
        tof.write(next_fileline)
        next_fileline = fromf.readline()
      else:
        tof.write(lines[lines_ind])
        tof.write("\n")
        lines_ind += 1
  util.mv_file(tof.name, filename)


if __name__ == "__main__":
  if len(sys.argv) <= 1 or "-h" in sys.argv:
    print(usage_str)
    exit(0)
  lines = [l[:-1] for l in sys.stdin.readlines()]  # remove newlines
  if sys.argv[1] == "-r":
    if len(sys.argv) <= 2:
      print(usage_str)
      exit(0)
    do(sys.argv[2], lines, reverse=True)
  else:
    do(sys.argv[1], lines)




####
# bomail.guistuff.mygui
#
# The main UI object.
####

import curses
import os
import sys
import locale

import bomail.config.config as config
import bomail.config.guiconfig as guicfg

import bomail.util.search_opts as search_opts

import bomail.cli.mailfile as mailfile
import bomail.cli.search as search
import bomail.cli.process as process
import bomail.cli.check_sched as check_sched

import bomail.guistuff.display as display
import bomail.guistuff.get_txt as get_txt

from bomail.util.addr import AddrBook
from bomail.util.thread import ThreadMgr
from bomail.util.tags import TagMgr

from bomail.guistuff.acts import Acts
from bomail.guistuff.tabthread import TabThread
from bomail.guistuff.tabnothread import TabNoThread


default_tab_ind = 0
default_tab_searches = [
  "-open",
  "-draft",
  "-sent",
  ""]



class Gui:
  def __init__(self, screen):
    self.init_screen(screen)

    self.mail_mgr = mailfile.MailMgr()
    self.addr_book = AddrBook()
    if guicfg.threads_on:
      self.thread_mgr = ThreadMgr()
    else:
      self.thread_mgr = None
    self.tag_mgr = TagMgr()
    self.acts = Acts(self)

    self.unread_set = set()  # unread filenames

    try:
      with open(config.tab_config_file) as f:
        s = f.read()
        namespace = {}
        exec(s, namespace)
        self.tab_ind = namespace["tab_ind"]
        tab_searches = namespace["tab_searches"]
    except:
      self.tab_ind = default_tab_ind
      tab_searches = default_tab_searches
    if guicfg.threads_on:
      self.tabs = [TabThread(s, self) for s in tab_searches]
    else:
      self.tabs = [TabNoThread(s, self) for s in tab_searches]

    self.tabs[self.tab_ind].load()  # better load initial tab now, during loading screen
    curses.flushinp()


  def init_screen(self, screen):
    self.screen = screen
    curses.curs_set(False)  # no cursor
    screen.scrollok(1)      # writing to last line/char is allowed
    curses.init_pair(guicfg.DEFAULT_CLR_PAIR, guicfg.foreground_color, guicfg.background_color)
    curses.init_pair(guicfg.BOMAIL_CLR_PAIR, guicfg.bomail_color, guicfg.background_color)
    curses.init_pair(guicfg.AUTHOR_CLR_PAIR, guicfg.author_color, guicfg.background_color)
    curses.init_pair(guicfg.TAGS_CLR_PAIR, guicfg.tags_color, guicfg.background_color)
    curses.init_pair(guicfg.THREAD_CLR_PAIR, guicfg.thread_color, guicfg.background_color)
      
    screen.keypad(True)
    
    self.attr = curses.color_pair(guicfg.BOMAIL_CLR_PAIR) | guicfg.bomail_attr
    self.screen.bkgd(" ", curses.color_pair(guicfg.BOMAIL_CLR_PAIR))
    self.tab_area = curses.newwin(curses.LINES - display.TOPLINES - display.BOTTOMLINES, curses.COLS, display.TOPLINES, 0)
    self.tab_area.bkgd(" ", curses.color_pair(guicfg.DEFAULT_CLR_PAIR))
    self.note_area = curses.newwin(1, curses.COLS, curses.LINES - display.BOTTOMLINES + 1, 0)
    self.note_area.bkgd(" ", curses.color_pair(guicfg.DEFAULT_CLR_PAIR))
    self.commands_area = curses.newwin(display.BOTTOMLINES-3, curses.COLS, curses.LINES - display.BOTTOMLINES + 3, 0)
    self.commands_area.bkgd(" ", curses.color_pair(guicfg.DEFAULT_CLR_PAIR))

    display.draw_loading_screen(self)


  def update_for_add(self, filelist, disp_infos):
    if guicfg.threads_on:
      self.thread_mgr.update_for_add(filelist, self.mail_mgr)
    for i,t in enumerate(self.tabs):
      t.update_for_add(filelist, disp_infos[i])

  def update_for_change(self, filelist, disp_infos):
    for i,t in enumerate(self.tabs):
      t.update_for_change(filelist, disp_infos[i])

  # filelist are the now-deleted filenames
  # data_list are their data (note they are no longer in mail manager!)
  def update_for_trash(self, filelist, data_list, disp_infos):
    fileset = set(filelist)
    if guicfg.threads_on:
      self.thread_mgr.update_for_trash(filelist, data_list, self.mail_mgr)
    for i,t in enumerate(self.tabs):
      t.update_for_trash(filelist, data_list, disp_infos[i])

  def rewrite_tab_searches(self):
    with open(config.tab_config_file, "w") as f:
      f.write("tab_ind = " + str(self.tab_ind) + "\n\n")
      f.write("tab_searches = [\n")
      for t in self.tabs:
        f.write("  \"" + t.search_str.replace('"','\\"') + "\",\n")
      f.write("]\n")

  def change_tab_ind(self, new_ind):
    self.tab_ind = new_ind

  def mark_read(self, filelist):
    for f in filelist:
      self.unread_set.discard(f)  # removes if present

  def mark_unread(self, filelist):
    self.unread_set.update(filelist)

  def go(self):
    keep_alive = True
    mode, note = "all", "Welcome"  # mode is all, tab, or note
    while keep_alive:
      display.redraw(self, mode, note)
      mode, note = "note", ""
      assert(all([(not t.is_loaded) or len(t.file_data) == 0 or (t.cursor_ind >= 0 and t.cursor_ind < len(t.file_data)) for t in self.tabs]))
      key = self.screen.getkey()
      keep_alive, used_key, mode, note = self.process_key(key)
      if not used_key:
        mode, note = self.tabs[self.tab_ind].process_key(key)
    # done with program; save tab settings and quit
    self.rewrite_tab_searches()
    return self.tabs[self.tab_ind].get_curr_filenames()

  # return 4 results:
  # 1) True if keep program alive, False to exit
  # 2) True if we processed the keypress, False otherwise
  # 3) if we processed, the mode to redraw
  # 4) if we processed, the note to draw
  def process_key(self, key):
    mode, note = "note", ""
    if key == guicfg.QUIT_KEY:
      return False, True, "", ""

    # tab navigation
    elif key >= "0" and key <= "9":
      key_num = int(key)
      new_ind = key_num - 1 if key_num > 0 else 9   # key "0" goes to the 10th tab
      if new_ind < len(self.tabs):
        self.change_tab_ind(new_ind)
        mode, note = "all", ""
    elif key == guicfg.TAB_LEFT_KEY:
      if self.tab_ind > 0:
        self.change_tab_ind(self.tab_ind - 1)
        mode, note = "all", ""
    elif key == guicfg.TAB_RIGHT_KEY:
      if self.tab_ind < len(self.tabs) - 1:
        self.change_tab_ind(self.tab_ind + 1)
        mode, note = "all", ""

    # editing tabs
    elif key == guicfg.EDIT_TAB_KEY:
      old_str = self.tabs[self.tab_ind].search_str
      new_str = get_txt.get_tab_search_txt(self, old_str)
      if new_str is None:
        mode, note = "all", "Cancelled"
      else:
        mode, note = self.acts.do(("edit tab", self.tab_ind, new_str, old_str))
    elif key == guicfg.SEARCH_IN_TAB_KEY:
      old_str = self.tabs[self.tab_ind].search_str
      q = get_txt.get_search_in_tab_txt(self)
      if q is None or len(q) == 0:
        mode, note = "all", "Cancelled"
      else:
        new_str = search_opts.get_new_search_str(old_str, q)
        mode, note = self.acts.do(("add tab", self.tab_ind + 1, new_str))
    elif key == guicfg.ADD_TAB_KEY:
      new_str = get_txt.get_tab_search_txt(self, "")
      if new_str is None:
        mode, note = "all", "Cancelled"
      else:
        mode, note = self.acts.do(("add tab", self.tab_ind + 1, new_str))
    elif key == guicfg.REMOVE_TAB_KEY:
      if len(self.tabs) > 0:
        mode, note = self.acts.do(("remove tab", self.tab_ind, self.tabs[self.tab_ind], self.tabs[self.tab_ind].get_old_display_info()))
      else:
        mode, note = "note", "Could not remove the only tab."
    elif key == guicfg.MOVE_TAB_LEFT_KEY:
      if self.tab_ind > 0:
        mode, note = self.acts.do(("move tab", self.tab_ind, -1))
    elif key == guicfg.MOVE_TAB_RIGHT_KEY:
      if self.tab_ind < len(self.tabs) - 1:
        mode, note = self.acts.do(("move tab", self.tab_ind, 1))

    # undo/redo
    elif key == guicfg.UNDO_KEY:
      if len(self.acts.acts) > 0 and self.acts.act_ind >= 0:
        mode, note = self.acts.undo()
      else:
        mode, note = "note", "Nothing to undo"
    elif key == guicfg.REDO_KEY:
      if len(self.acts.acts) > 0 and self.acts.act_ind + 1 < len(self.acts.acts):
        mode, note = self.acts.redo()
      else:
        mode, note = "note", "Nothing to redo"

    elif key == guicfg.GET_KEY:  # get updates
        disp_infos = [t.get_old_display_info() for t in self.tabs]
        display.redraw_note(self, "Getting new mail and updates (please wait)...")
        new_files = process.main(self.mail_mgr, self.tag_mgr)
        unsched_files = check_sched.main(self.mail_mgr)
        self.update_for_add(new_files, disp_infos)
        self.update_for_change(unsched_files, disp_infos)
        self.mark_unread(new_files)
        self.mark_unread(unsched_files)
        curses.flushinp()  # it might've been a long wait, flush keypresses
        unsched_str = "" if len(unsched_files) == 0 else ", de-scheduled " + str(len(unsched_files)) + " emails"
        mode, note = "all", "Got " + str(len(new_files)) + " new emails" + unsched_str

    else:  # if none of the above keys match, send the keypress to the tab
      return True, False, "", ""

    # if one of the keys matched, we fell through to here
    return True, True, mode, note


def main(screen):
  filenames = Gui(screen).go()
  curses.endwin()
  os.system("clear")
  print("\n".join(filenames))




###
# bomail.guistuff.acts
#
# "acts" are any undoable actions taken by the program,
# such as changing the state of an email, adding tags, or gui actions.
####

import sys
import dateutil.parser

import bomail.config.guiconfig as guicfg
import bomail.config.config as config

import bomail.cli.search as search
import bomail.cli.chstate as chstate
import bomail.cli.mailfile as mailfile
import bomail.util.datestuff as datestuff
import bomail.guistuff.display as display
import bomail.guistuff.display_fmt as display_fmt
import bomail.guistuff.tabnothread as tabnothread
import bomail.guistuff.tabthread as tabthread


class Acts:
  def __init__(self, gui):
    self.acts = []
    self.act_ind = -1
    self.gui = gui

  def do(self, act):
    self.acts = self.acts[:self.act_ind + 1]  # cut off any old branches
    self.act_ind = len(self.acts) - 1         # just in case the index got too large
    self.acts.append(act)
    return self.redo()

  # return pair a, note
  # where a = kind of redraw to do
  def redo(self):
    self.act_ind += 1
    act = self.acts[self.act_ind]
    return self.exec_act(act)
    
  def exec_act(self, act):
    with open(config.acts_log_file, "a") as f:
      f.write("DO: " + str(act) + "\n")
    if act[0] in ["trash", "open", "closed", "scheduled", "add tags", "remove tags", "add attach", "remove attach"]:
      if len(act[1]) >= 100:  # number 100 chosen via extensive UI studies
        display.redraw_note(self.gui, "Processing command (please wait)...")

    if act[0] == "trash":  # ("trash", filelist, datalist)
      shorts = ", ".join([display_fmt.get_shortened(d, self.gui.mail_mgr) for d in act[1]])
      disp_infos = [t.get_old_display_info() for t in self.gui.tabs]
      chstate.trash(act[1], self.gui.mail_mgr)
      self.gui.update_for_trash(act[1], act[2], disp_infos)
      return "all", "Trashed " + shorts

    elif act[0] in ["open", "closed", "scheduled"]:  # (act, filelist, prevstatelist [, schedobj])
      shorts = ", ".join([display_fmt.get_shortened(d, self.gui.mail_mgr) for d in act[1]])
      disp_infos = [t.get_old_display_info() for t in self.gui.tabs]
      if act[0] == "open":
        chstate.make_open(act[1], self.gui.mail_mgr)
        note = "Opened " + shorts
      elif act[0] == "closed":
        chstate.make_closed(act[1], self.gui.mail_mgr)
        note = "Closed " + shorts
      else:
        chstate.schedule(act[1], act[3], self.gui.mail_mgr)
        note = "Scheduled for " + datestuff.get_local_str(act[3]).replace("T"," ") + ": " + shorts
      self.gui.update_for_change(act[1], disp_infos)
      return "all", note

    elif act[0] == "add tags":  # ("add tags", filelist, addedtagslist, oldtagslists, brandnewtags)
      shorts = ", ".join([display_fmt.get_shortened(d, self.gui.mail_mgr) for d in act[1]])
      disp_infos = [t.get_old_display_info() for t in self.gui.tabs]
      self.gui.mail_mgr.add_tags(act[1], act[2])
      if len(act[4]) > 0:
        self.gui.tag_mgr.add_tags(act[4])
      self.gui.update_for_change(act[1], disp_infos)
      return "all", "Added tags: " + ", ".join(act[2]) + " to " + shorts

    elif act[0] == "remove tags":  # ("remove tags", filelist, removedtagslist, oldtagslists)
      disp_infos = [t.get_old_display_info() for t in self.gui.tabs]
      shorts = ", ".join([display_fmt.get_shortened(d, self.gui.mail_mgr) for d in act[1]])
      self.gui.mail_mgr.remove_tags(act[1], act[2])
      self.gui.update_for_change(act[1], disp_infos)
      return "all", "Removed tags: " + ", ".join(act[2]) + " from " + shorts

    elif act[0] == "add attach":  # ("add attach", filelist, old_attachlist, new_attachlist)
      shorts = ", ".join([display_fmt.get_shortened(d, self.gui.mail_mgr) for d in act[1]])
      disp_infos = [t.get_old_display_info() for t in self.gui.tabs]
      for f in act[1]:
        self.gui.mail_mgr.set_attachlist(f, act[2] + act[3])
      self.gui.update_for_change(act[1], disp_infos)
      return "tab", "Added attachment " + ", ".join(act[3]) + " to " + shorts

    elif act[0] == "remove attach":  # ("remove attach", filelist, prev_attachlist, remove_attachlist)
      shorts = ", ".join([display_fmt.get_shortened(d, self.gui.mail_mgr) for d in act[1]])
      disp_infos = [t.get_old_display_info() for t in self.gui.tabs]
      for f in act[1]:
        self.gui.mail_mgr.set_attachlist(f, [s for s in act[2] if s not in act[3]])
      self.gui.update_for_change(act[1], disp_infos)
      return "tab", "Removed attachment " + ", ".join(act[3]) + " from " + shorts

    elif act[0] == "edit tab":  # ("edit tab", tab_ind, new_str, old_str)
      if not search.is_valid_searchstr(act[2]):
        return "all", "Cancelled, not a valid search string: " + act[2]
      self.gui.tabs[act[1]].search_str = act[2]
      self.gui.tabs[act[1]].load()
      self.gui.rewrite_tab_searches()
      return "all", "Changed tab search string to: " + act[2]

    elif act[0] == "add tab":  # ("add tab", tab_ind, new_str)
      if not search.is_valid_searchstr(act[2]):
        return "all", "Cancelled, not a valid search string: " + act[2]
      if guicfg.threads_on:
        self.gui.tabs.insert(act[1], tabthread.TabThread(act[2], self.gui))
      else:
        self.gui.tabs.insert(act[1], tabnothread.TabNoThread(act[2], self.gui))
      self.gui.change_tab_ind(act[1])
      self.gui.rewrite_tab_searches()
      return "all", "Added new tab with search string: " + act[2]

    elif act[0] == "remove tab":  # ("remove tab", tab_ind, tab object, disp_info)
      del self.gui.tabs[act[1]]
      self.gui.change_tab_ind(min(act[1]-1, len(self.gui.tabs)-1))
      self.gui.rewrite_tab_searches()
      return "all", "Removed tab with search string: " + act[2].search_str

    elif act[0] == "move tab":  # ("move tab", tab_ind, amount)
      src = act[1]
      dest = src + act[2]
      self.gui.tabs[src], self.gui.tabs[dest] = self.gui.tabs[dest], self.gui.tabs[src]
      self.gui.change_tab_ind(dest)
      self.gui.rewrite_tab_searches()
      return "all", "Shifted tab " + ("left" if dest < src else "right")

    elif act[0] == "edit file":  # ("edit file", filename, mailfile_line, new_str, old_str)
      self.gui.mail_mgr.set(act[1], act[2], act[3])
      return "all", "Edited [" + mailfile.fields[act[2]] + "] of " + display_fmt.get_shortened(act[1], self.gui.mail_mgr) + " to: " + act[3]

    else:
      return "all", "bomail error: unknown action: " + act[0]


  def undo(self):
    act = self.acts[self.act_ind]
    self.act_ind -= 1
    with open(config.acts_log_file, "a") as f:
      f.write("UNDO: " + str(act) + "\n")

    if act[0] in ["trash", "open", "closed", "scheduled"]:
      if act[0] == "trash":  # ("trash", filelist, datalist)
        chstate.untrash(act[1], act[2], self.gui.mail_mgr)
        prevstates = [d[mailfile.STATE_L] for d in act[2]]
      else:  # (act, filelist, prevstatelist [, sched])
        prevstates = act[2]
      shorts = ", ".join([display_fmt.get_shortened(d, self.gui.mail_mgr) for d in act[1]])
      
      openlist = []
      closedlist = []
      disp_infos = [t.get_old_display_info() for t in self.gui.tabs]
      for f,s in zip(act[1], prevstates):
        if s == "open":
          openlist.append(f)
        elif s == "closed":
          closedlist.append(f)
        else:
          chstate.schedule([f], dateutil.parser.parse(s.split()[1]), self.gui.mail_mgr)
      if len(openlist) > 0:
        chstate.make_open(openlist, self.gui.mail_mgr)
      if len(closedlist) > 0:
        chstate.make_closed(closedlist, self.gui.mail_mgr)
      if act[0] == "trash":
        self.gui.update_for_add(act[1], disp_infos)
      else:
        self.gui.update_for_change(act[1], disp_infos)
      return "all", "Undid change of state to " + act[0] + " for " + shorts

    elif act[0] in ["add tags", "remove tags"]:  # (act, filelist, tagslist, oldtagslists [, brand_new_tags])
      shorts = ", ".join([display_fmt.get_shortened(d, self.gui.mail_mgr) for d in act[1]])
      disp_infos = [t.get_old_display_info() for t in self.gui.tabs]
      for f,prev in zip(act[1], act[3]):
        self.gui.mail_mgr.set_tags([f], prev)
      if act[0] == "add tags" and len(act[4]) > 0:
        self.gui.tag_mgr.remove_tags(act[4])
      self.gui.update_for_change(act[1], disp_infos)
      return "all", "Undid edit of tags for " + shorts

    elif act[0] == "add attach":  # ("add attach", filelist, old_attachlist, new_attachlist)
      shorts = ", ".join([display_fmt.get_shortened(d, self.gui.mail_mgr) for d in act[1]])
      disp_infos = [t.get_old_display_info() for t in self.gui.tabs]
      for f in act[1]:
        self.gui.mail_mgr.set_attachlist(f, act[2])
      self.gui.update_for_change(act[1], disp_infos)
      return "tab", "Undid new attachment to " + shorts

    elif act[0] == "remove attach":  # ("remove attach", filelist, prev_attachlist, remove_attachlist)
      shorts = ", ".join([display_fmt.get_shortened(d, self.gui.mail_mgr) for d in act[1]])
      disp_infos = [t.get_old_display_info() for t in self.gui.tabs]
      for f in act[1]:
        self.gui.mail_mgr.set_attachlist(f, act[2])
      self.gui.update_for_change(act[1], disp_infos)
      return "tab", "Undid removal of attachment from " + shorts

    elif act[0] == "edit file":  # ("edit file", filename, mailfile_line, new_str, old_str)
      self.gui.mail_mgr.set(act[1], act[2], act[4])
      return "all", "Undid edit of [" + mailfile.fields[act[2]] + "] for " + display_fmt.get_shortened(act[1], self.gui.mail_mgr)

    elif act[0] == "edit tab":  # ("edit tab", tab_ind, new_str, old_str)
      self.gui.tabs[act[1]].search_str = act[3]
      self.gui.tabs[act[1]].is_loaded = False
      self.gui.rewrite_tab_searches()
      return "all", "Undid change of tab search string"

    elif act[0] == "add tab":  # ("add tab", tab_ind, new_str)
      del self.gui.tabs[act[1]]
      self.gui.change_tab_ind(max(0, act[1]-1))
      self.gui.rewrite_tab_searches()
      return "all", "Undid addition of new tab"

    elif act[0] == "remove tab":  # ("remove tab", tab_ind, tab_object, disp_info)
      self.gui.tabs.insert(act[1], act[2])
      act[2].recheck(act[3])  # for changes since it was last used/updated
      self.gui.change_tab_ind(act[1])
      self.gui.rewrite_tab_searches()
      return "all", "Undid removal of tab"

    elif act[0] == "move tab":  # ("move tab", tab_ind, amount)
      dest = act[1]
      src = dest + act[2]
      self.gui.tabs[src], self.gui.tabs[dest] = self.gui.tabs[dest], self.gui.tabs[src]
      self.gui.change_tab_ind(dest)
      self.gui.rewrite_tab_searches()
      return "all", "Undid shift of tab"




####
# bomail.util.unpack_attach
#
# Unpack attachments from a MIME file into a directory
####

import sys
import os, email

allowed_chars = (' ','.','_')


def sanitize(filename):
  return "".join(c for c in filename if c.isalnum() or c in allowed_chars).rstrip()

def get_fname(base_filename):
  filename = base_filename
  num = 0
  while os.path.exists(filename):
    num += 1
    filename = base_filename + "_" + str(num)
  return filename


def is_attachment(part):
  cd = part.get_content_disposition()
  return cd is not None and cd.lower() == "attachment"

  
# return a list of attachment file paths
def do_unpack(py_msg, attach_dir):
  made_dir = False
  results = []
  for part in py_msg.walk():
    if is_attachment(part):
      unclean_filename = part.get_filename()
      if unclean_filename is None:
        unclean_filename = "unknownfile"
      filename = get_fname(attach_dir + sanitize(unclean_filename))
      try:
        payload = part.get_payload(decode=True)
        mode = "wb"
      except:
        # "This should never happen" thanks python
        payload = part.get_payload(decode=False)
        mode = "w"
      if payload is not None:
        if not made_dir:
          os.makedirs(attach_dir, exist_ok=True)
          made_dir = True
        with open(filename, mode) as f:
          f.write(payload)
        results.append(filename)
  return results




####
# bomail.util.thread
#
# Manages EMAIL threads (not multithreading), a.k.a. conversations.
####

import sys
import os
import bisect

import bomail.config.config as config
import bomail.cli.mailfile as mailfile
import bomail.util.util as util
import bomail.util.msgids as msgids
import bomail.util.datestuff as datestuff
import bomail.guistuff.display_fmt as display_fmt

####
# source of algorithm:
# http://www.jwz.org/doc/threading.html 

# Every possible email has a unique msg_id;
# some of them also have filenames present.
# For every msg_id we hear about that is relevant, we have a "container"
# and entry in id_to_cntr.
# Each thread is a tree of containers with some root.
# Threads also have "pruned" versions that remove containers for which
# we don't have filenames.
####

usage_str = """
Usage:
    python3 thread.py -h     # print this help
    python3 thread.py

Reads a list of files from stdin, one per line.
Prints a representation of each thread those files belong to.
"""

class ThreadMgr:
  def __init__(self):
    self.id_to_cntr = {}         # msg_id to its Container
    self.id_to_pruned_root = {}  # msg_id to its root's Container

  def update_for_add(self, filelist, mail_mgr):
    touched_cntrs = build_id_to_cntr(filelist, mail_mgr, self.id_to_cntr)
    self.do_pruning(touched_cntrs, mail_mgr)

  def do_pruning(self, touched_cntrs, mail_mgr):
    touched_rootset = set()
    for c in touched_cntrs:
      while c.parent is not None:
        c = c.parent
      touched_rootset.add(c)
    pruned_roots = prune_empty(list(touched_rootset), mail_mgr, is_root_list=True)
    for r in pruned_roots:
      r.point_to_pruned_root(r, self.id_to_pruned_root)
      r.recount()

  # don't remove container, because its msg_id may be part of a useful chain
  # and don't look up any of these files in mail_mgr, as they don't exist anymore!
  def update_for_trash(self, filelist, data_list, mail_mgr):
    msg_id_list = [d[mailfile.MSG_ID_L] for d in data_list]
    touched_cntrs = []
    for msg_id in msg_id_list:
      if msg_id in self.id_to_cntr:
        c = self.id_to_cntr[msg_id]
        c.remove_self()
        touched_cntrs.append(c)
    self.do_pruning(touched_cntrs, mail_mgr)


  # given a "base" filelist
  # Return a list of triples (matching_files, root_cntr, all_files) where:
  #  - each triple corresponds to one thread
  #  - matching_files are those from the *base* list that are in this thread
  #  - root_cntr is the Container of the root of the thread
  #  - all_files is a list in some order of *all* files in thread
  # sort order of matching_files and the returned list is according to filelist
  def get_threads_for(self, filelist, mail_mgr):
    if len(filelist) == 0:
      return []
    msg_id_list = [mail_mgr.get(f, mailfile.MSG_ID_L) for f in filelist]
    new_flist = [f for m,f in zip(msg_id_list,filelist) if m not in self.id_to_pruned_root]
    self.update_for_add(new_flist, mail_mgr)
    return self.get_the_threads(filelist, msg_id_list)


  # this is pretty slow with many threads...
  # return list of triples (matching_files_in_thread, root, all_files_in_thread)
  def get_the_threads(self, filelist, msg_id_list):
    results = []        # list of pairs [files_in_this_thread, root_container]
    results_roots = {}  # map container to int, its index in results
    for msg_id,f in zip(msg_id_list, filelist):
      if msg_id not in self.id_to_pruned_root:
        with open(config.error_log_file, "a") as f_err:
          f_err.write(datestuff.get_local_nowstr() + "\n")
          f_err.write("Thread error: message id not in self.id_to_pruned_root.\n" + str(msg_id) + "\n" + str(f) + "\nIs in id_to_cntr: " + str(msg_id in self.id_to_cntr) + "\n\n")
        continue
      r = self.id_to_pruned_root[msg_id]
      if r in results_roots:
        results[results_roots[r]][0].append(f)
      else:
        temp = []
        r.build_fname_list(temp)
        results_roots[r] = len(results)
        results.append(([f], r, [t[1] for t in temp]))
    return results


# actually seems never necessary to do this
#def sort_msgs(container_list, mgr):
#  def cntr_to_val(c):
#    return (("9999-12-31" if c.filename is None else mgr.get(c.filename, mailfile.DATE_L)) , c.msg_id)
#  container_list.sort(key=cntr_to_val)


# MAIN THREAD-BUILDING ALGORITHM
# build all message Containers and put in the table id_to_cntr
# return list of all files touched
def build_id_to_cntr(orig_filelist, mgr, id_to_cntr):
  filelist = list(orig_filelist)
  fileset = set(filelist)
  file_i = 0
  touched_cntrs = []

  # go through files creating containers for all msg_ids encountered
  # and setting their parent-child relationships
  while file_i < len(filelist):  # note filelist gets lengthened dynamically
    filename = filelist[file_i]
    msg_id = mgr.get(filename, mailfile.MSG_ID_L)

    # A. get this message's container
    if msg_id in id_to_cntr:
      cntr = id_to_cntr[msg_id]
    else:
      cntr = Container(msg_id)
      id_to_cntr[msg_id] = cntr
    cntr.filename = filename
    touched_cntrs.append(cntr)

    # A.5 (Bo's modification) append my "children" messages to filelist if not there
    children_ids = mgr.get_refby(filename)
    for c_id in children_ids:
      if c_id in mgr.ids.ids_to_filenames and c_id not in id_to_cntr:
        c_fname = mgr.ids.get(c_id)
        if c_fname not in fileset:
          fileset.add(c_fname)
          filelist.append(c_fname)

    # B. link its references list together as "parent of parent of ... of this"
    # First create containers for any of them if necessary
    # Also, add their files to filelist if not present
    ref_ids = mgr.get_references(filename)
    ref_containers = []
    for ref_id in ref_ids:
      if ref_id == mgr.get(filename, mailfile.MSG_ID_L):  # this happens sometimes
        continue

      # add its containers to table if not present
      if ref_id in id_to_cntr:
        ref_cntr = id_to_cntr[ref_id]
      else:
        ref_cntr = Container(ref_id)
        id_to_cntr[ref_id] = ref_cntr
      ref_containers.append(ref_cntr)
      # also append ref_id's file onto filelist if it exists
      if ref_id in mgr.ids.ids_to_filenames:
        ref_filename = mgr.ids.ids_to_filenames[ref_id]
        if ref_filename not in fileset:
          filelist.append(ref_filename)
          fileset.add(ref_filename)
    # Second, link the containers, but don't overwrite existing parent-child links
    for i in range(len(ref_containers)-1):
      # do the linking
      a,b = ref_containers[i], ref_containers[i+1]
      if b.parent is None and not a.has_descendant(b.msg_id) and not b.has_descendant(a.msg_id):
        a.add_child(b, mgr)

    # C. definitively set the parent of this message, overwriting existing relationships
    if cntr.parent is not None:
      cntr.parent.children.remove(cntr)
      cntr.parent = None
    if len(ref_containers) > 0:
      # This if-condition failing would be very bizarre, but it's happened to me
      if not cntr.has_descendant(ref_containers[-1].msg_id):
        ref_containers[-1].add_child(cntr, mgr)

    # D. increment file_i
    file_i += 1
  return touched_cntrs


# Prune empty containers (those with no file) where possible.
# Originally given list of root containers, recursively works down.
# If given container is empty, prune it if:
#  - it's a root with only one child - the child takes its place.
#  - it's a non-root - all its children get "spliced" in at its level.
# Return the pruned list (do not modify original)
def prune_empty(cntr_list, mail_mgr, is_root_list=False):
  if len(cntr_list) == 0:
    return []

  # A. prune all children of containers in this list
  for c in cntr_list:
    c.pruned_children = prune_empty(c.children, mail_mgr)

  # B. prune empty containers from this list and promote their children
  new_cntr_list = []
  for c in cntr_list:
    if c.filename is None and (not is_root_list or len(c.pruned_children) == 1):
      new_cntr_list += c.pruned_children
    else:
      new_cntr_list.append(c)

#  sort_msgs(new_cntr_list, mail_mgr)
  return new_cntr_list
  

class Container:
  def __init__(self, msg_id):
    self.msg_id = msg_id
    self.parent = None  # Container
    self.children = []
    self.pruned_children = []   # prunes msg_ids with no associated file
    self.filename = None
    self.count = 0  # number in descendents (including self) who have a filename


  def __str__(self):
    return "Container " + self.msg_id + " : " + str(self.filename) + " : " + str(self.children)

  def __repr__(self):
    return str(self)

  def __hash__(self):
    return hash(self.msg_id)

  def __eq__(self, other):
    return self.msg_id == other.msg_id

  def __ne__(self, other):
    return self.msg_id != other.msg_id


  def add_child(self, child, mgr):
    self.children.append(child)
    child.parent = self
#    sort_msgs(self.children, mgr)


  def get_root(self):
    return self if self.parent is None else self.parent.get_root()


  # Do NOT use pruned_children for this, or we might miss a relationship
  # that has been pruned
  def has_descendant(self, mid):
    if self.msg_id == mid:
      return True
    for c in self.children:
      if c.has_descendant(mid):
        return True
    return False

  def remove_self(self):
    self.filename = None


  # point pruned msg_ids to roots
  def point_to_pruned_root(self, r, id_to_pruned_root):
    id_to_pruned_root[self.msg_id] = r
    for c in self.pruned_children:
      c.point_to_pruned_root(r, id_to_pruned_root)


  def recount(self):
    num = 0 if self.filename is None else 1
    for c in self.pruned_children:
      num += c.recount()
    self.count = num
    return num


  # add (level, filename) to list in preorder depth-first order:
  # me, first child and all descendants, second child and all descendants, ....
  def build_fname_list(self, a, level=0):
    if self.filename is not None:
      a.append((level, self.filename))
    for c in self.pruned_children:
      c.build_fname_list(a, level+1)


if __name__ == "__main__":
  if "-h" in sys.argv:
    print(usage_str)
    exit(0)
  width = 100  # TODO command-line option

  mail_mgr = mailfile.MailMgr()
  thread_mgr = ThreadMgr()
  filelist = [f.strip() for f in sys.stdin.readlines()]
  thread_trips = thread_mgr.get_threads_for(filelist, mail_mgr)
  for trip in thread_trips:
    print("================")
    for f in trip[2]:  # all files in thread
      print(("--> " if f in trip[0] else "    ") + f)
      # TODO: sort and indent according to thread replies, depending on command-line option
      lines, blank_attr_data = display_fmt.get_msg_lines_nothread(mail_mgr, f, width, use_curses=False)
      for l in lines:
        print("    " + l)
      print()
    print()





####
# bomail.cli.stats
#
# Report some stats about a set of mail filenames.
####

import sys
import string
import datetime
import dateutil.parser
from dateutil.relativedelta import relativedelta
import os
import shlex
import io
import math

import bomail.cli.mailfile as mailfile
import bomail.cli.search as search
import bomail.util.addr as addr
import time

# TODO: implement line wrapping for all printing

iso_wkdays = ['Mon','Tue','Wed','Thu','Fri','Sat','Sun']
month_names = ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec']

max_items = 25
max_cols = 100  # width to print


usage_str = """
Usage:
    bomail stats [search_str]  # run bomail stats on all emails matching search_str

Global options:
    bomail stats -max_items n  # print at most n items per list (default """ + str(max_items) + """)
    bomail stats -max_cols m   # wrap lines at m characters
"""


help_str = """
------------------------------------------
1. Printing basic stats:
    top [list]       # print top entries ever in list, aggregated
    num [list]       # print statistics about length of list

Available lists:
  - addresses
  - authors
  - recipients
  - tags
  - words
  - attach

Examples:
    top recipients   # get most common recipients
    num recipients   # get stats about num of recips per message
    top words        # get most commonly used words
    num words        # get stats about num of words per message


------------------------------------------
2. Filter using a search string:
    matching [search_str]             # print default stats for matching files
    total matching [search_str]       # how many messages match the query
    fraction matching [search_str]    # what fraction of all messages match the query
    top [list] matching [search_str]
    num [list] matching [search_str]

Examples:
    matching -subject "Hungry"        # default stats for emails with "Hungry" subj
    top words matching -sent          # what words do I like to use?
    num words matching -sent          # how long are emails I write?
    fraction matching -sent           # fraction of emails sent by me
    top authors matching "chocolate"  # list those who write about it most


------------------------------------------
3a. Slicing into bins:
    slice [basic_query] into [bin_units]
    slice [filter_query] into [bin_units]

Bin units:
  - years
  - months
  - weeks
  - days
  - hours

Examples:
    slice into years                            # stats about number of messages each year
    slice top addresses into months             # list top addresses of Jan, Feb, ...
    slice num addresses into months             # plot avg num addresses per message, each month
    slice fraction matching -sent into months
    slice num tags matching -attach into weeks


3b. Slicing across time:
    slice [basic_part] into [bin_units] of the [slice_unit]
    slice [filter+part] into [bin_units] of the [slice_unit]

Same thing, but group together bins across different slice_units:
  - year
  - month
  - week
  - day

Examples:
    slice total into days of the month
    slice fraction matching -from me -to "Name" into days of the week

The first gave how many emails I get on the 1st, 2nd, 3rd, ... of each month.
The second gave what fraction of all emails I send on Mon, Tue, ... go to "Name".


"""


def datestr_to_week_of_year(datestr):
  return datetime.date(int(datestr[:4]), int(datestr[5:7]), int(datestr[8:10])).isocalendar()[1]


def itemlist_to_strlist(itemlist, item_to_str):
  res = []
  templist = itemlist[:max_items]
  for i,pr in enumerate(templist):
    res.append("    ")
    res.append(item_to_str(pr[0]))
    res.append(" (")
    res.append(str(pr[1]))
    res.append(")")
    if i < len(templist) - 1:
      res.append("\n")
  return res


# given a sorted list of numbers
def compute_stats(counts):
  n = len(counts)
  if n == 0:
    return 0, 0, 0, 0, 0, 0, 0
  mini = min(counts)
  maxi = max(counts)
  avg = sum(counts) / n
  std_dev = -1 if n==1 else math.sqrt(sum([(c-avg)**2 for c in counts]) / (n-1))
  counts.sort()
  quant_05 = counts[int(n/20)]
  quant_25 = counts[int(n/4)]
  median = counts[int(n/2)]
  quant_75 = counts[int(3*n/4)]
  quant_95 = counts[int(19*n/20)]
  return sum(counts), mini, maxi, avg, std_dev, quant_05, quant_25, median, quant_75, quant_95


def get_stats_strlist_from_counts(counts):
  return get_stats_strlist(compute_stats(counts))


def get_stats_strlist(stats):
  total, mini, maxi, avg, std_dev, quant_05, quant_25, median, quant_75, quant_95 = stats
  return ["    total:   ", str(total), "\n",
          "    min:     ", str(mini), "\n",
          "    max:     ", str(maxi), "\n",
          "    average: ", str(avg), "\n",
          "    std dev: ", str(std_dev), "\n",
          "     5 %:    ", str(quant_05), "\n",
          "    25 %:    ", str(quant_25), "\n",
          "    median:  ", str(median), "\n",
          "    75 %:    ", str(quant_75), "\n",
          "    95 %:    ", str(quant_95), "\n",
          "\n"]


# helper functions for slicing over dates
# get the endpoint of the bin this date is in
def round_up(date_obj, bin_unit):
  # just to be safe
  if date_obj.year == 9999:
    return date_obj

  if bin_unit.startswith("hour"):
    upper = date_obj + datetime.timedelta(hours = 1)
    return datetime.datetime(year=upper.year, month=upper.month, day=upper.day, hour=upper.hour, tzinfo=upper.tzinfo)
  elif bin_unit.startswith("day"):
    upper = date_obj + datetime.timedelta(days = 1)
    return datetime.datetime(year=upper.year, month=upper.month, day=upper.day, tzinfo=upper.tzinfo)
  elif bin_unit.startswith("week"):
    wkday = date_obj.isocalendar()[2]
    upper = date_obj + datetime.timedelta(days = 7 - wkday)
    return datetime.datetime(year=upper.year, month=upper.month, day=upper.day, tzinfo=upper.tzinfo)
  elif bin_unit.startswith("month"):
    lower = datetime.datetime(year=date_obj.year, month=date_obj.month, day=1, tzinfo=date_obj.tzinfo)
    return lower + relativedelta(months=+1)
  elif bin_unit.startswith("year"):
    return datetime.datetime(year=min(9999,date_obj.year+1), month=1, day=1, tzinfo=date_obj.tzinfo)


def datestr_to_bin_id(datestr, bin_unit, slice_unit):
  if slice_unit is not None and slice_unit.startswith("week"):
    day_num = dateutil.parser.parse(datestr).isocalendar()[2] - 1
    wkday = iso_wkdays[day_num]
    if bin_unit.startswith("day"):
      return wkday
    # assume it's hours
    return wkday + " " + datestr[11:13]
  if bin_unit.startswith("week"):
    if slice_unit is None:
      return datestr[:4] + " " + str(datestr_to_week_of_year(datestr))
    if slice_unit.startswith("year"):
      return str(datestr_to_week_of_year(datestr))
    # assume it's month, and break the weeks of the month into these sets of days
    return ["01-07","08-14","15-21","22-28","29-31"][int(int(datestr[8:10]) / 7)]

  max_len = 13
  if bin_unit.startswith("day"):
    max_len = 10
  elif bin_unit.startswith("month"):
    if slice_unit is not None and slice_unit.startswith("year"):
      return month_names[int(datestr[5:7])-1]
    max_len = 7
  elif bin_unit.startswith("year"):
    max_len = 4

  if slice_unit is None:
    return datestr[:max_len].replace("T"," ")
  elif slice_unit.startswith("year"):
    return datestr[5:max_len].replace("T"," ")
  elif slice_unit.startswith("month"):
    return datestr[8:max_len].replace("T"," ")
  else:
    return datestr[11:max_len].replace("T"," ")


# return two functions which give the name of a date's bin and its slice
def get_date_maps(bin_unit, slice_unit):
  date_to_sliceind = lambda x: x  # map date to the name of the slice it's in
  date_to_binname = lambda x: x   # map date to the name of the bin it's in

  if slice_unit.startswith("year"):
    date_to_sliceind = lambda datestr: datestr[:4]
    if bin_unit.startswith("month"):
      date_to_binname = lambda datestr: datestr[5:7]
    elif bin_unit.startswith("week"):
      date_to_binname = datestr_to_week_of_year
    elif bin_unit.startswith("day"):
      date_to_binname = lambda datestr: datestr[5:10]
    else:  # hour
      date_to_binname = lambda datestr: datestr[5:]

  elif slice_unit.startswith("month"):
    date_to_sliceind = lambda datestr: datestr[:7]
    if bin_unit.startswith("week"):
      date_to_binname = lambda datestr: int(int(datestr[8:10]) / 7)
    elif bin_unit.startswith("day"):
      date_to_binname = lambda datestr: datestr[8:10]
    else:  # hour
      date_to_binname = lambda datestr: datestr[8:]

  elif slice_unit.startswith("week"):
    # (year, week of the year)
    date_to_sliceind = lambda datestr: (datestr[:4], dateutil.parser.parse(datestr).isocalendar()[1])
    get_wkday = lambda datestr: str(datetime.date(int(datestr[:4]), int(datestr[5:7]), int(datestr[8:10])).weekday())
    if bin_unit.startswith("day"):
      date_to_binname = get_wkday
    else:  # hour
      date_to_binname = lambda datestr: get_wkday(datestr) + " " + datestr[11:]

  else:  # day: assume hours
    date_to_sliceind = lambda datestr: datestr[:10]
    date_to_binname = lambda datestr: int(datestr[11:])

  return date_to_sliceind, date_to_binname


# binstr_to_flist: a dict mapping each binstr to list of files that fall in that bin
#   binstr is a rounded date representing a bin
# return list of tuples: (bin_identifier, bin_filelist, num_sliced_over)
# sorted by bins chronologically
# e.g. if we bin into months and slice across two years 1994-1995, then return a list of 12 tuples, each
# of the form e.g. ('Jan', [filelist], 2)
def do_slice(binstr_to_flist, bin_unit, slice_unit):
  if slice_unit is None:
    mylist = sorted(binstr_to_flist.items())
    return [(datestr_to_bin_id(d, bin_unit, slice_unit), flist, 1) for d,flist in mylist]

  date_to_sliceind, date_to_binname = get_date_maps(bin_unit, slice_unit)
  bins = {}  # map slice_ind to tuple (bin_id, [filelist], set(unique_slices)) 
  for datestr,flist in binstr_to_flist.items():
    k = date_to_binname(datestr)
    if k in bins:
      bins[k][1] += flist
      bins[k][2].add(date_to_sliceind(datestr))
    else:
      bins[k] = [datestr_to_bin_id(datestr, bin_unit, slice_unit), list(flist), set([date_to_sliceind(datestr)])]
  result = [(pr[1][0], pr[1][1], len(pr[1][2])) for pr in sorted(bins.items())]

  return result


# totals is a list of numbers
def get_bar_plot(labels, totals):
  s_nums = ['%s' % float('%.4g' % t) for t in totals]
  max_numlen = max(map(len,s_nums))
  max_lablen = max(map(len,labels))
  cols = max(1, max_cols - max_numlen - max_lablen - 4)
  m = max(totals)
  col_list = [round(b*cols/m) for b in totals]
  slist = []
  for i in range(len(totals)):
    slist.append(labels[i].ljust(max_lablen) + " (" + (s_nums[i] + ")").ljust(max_numlen+1) + " " + ("="*col_list[i]))
    slist.append("\n")
  return slist


def get_list_plot(labels, attrs, listname):
  max_lablen = max(map(len,labels))
  slist = []
  for i in range(len(attrs)):
    slist.append(labels[i].ljust(max_lablen) + "\n")
    slist += itemlist_to_strlist(attrs[i], get_strfunc(listname))
    slist.append("\n")
  return slist


def get_mean_plot(labels, stats):
  avgs = [s[3] for s in stats]
  std_devs = [s[4] for s in stats]
  s_nums = list(map(str, avgs))
  max_numlen = max(map(len,s_nums))
  max_lablen = max(map(len,labels))
  cols = max(1, max_cols - max_numlen - max_lablen - 4)
  m = max([a+s for a,s in zip(avgs,std_devs)])
  avg_list = [round(b*cols/m) for b in avgs]
  low_list = [max(0,round((a-s)*cols/m)) for a,s in zip(avgs,std_devs)]
  hi_list = [min(cols-1, round((a+s)*cols/m)) for a,s in zip(avgs,std_devs)]
  numstrs = [" "*l + "-"*(a-l) + "=" + "-"*(h-a) for a,l,h in zip(avg_list, low_list, hi_list)]
  slist = []
  for i in range(len(stats)):
    avg_str = '%s' % float('%.4g' % avgs[i])
    sd_str  = '%s' % float('%.4g' % std_devs[i])
    slist.append(labels[i].ljust(max_lablen) + " (" + (avg_str + " ± " + sd_str + ")").ljust(max_numlen+1) + " " + numstrs[i])
    slist.append("\n")
  return slist


# persistent, re-usable
punct_trans = str.maketrans("","", string.punctuation)

# a summary of a message
class MsgSum:
  def __init__(self, filename, mail_mgr):
    self.authlist = addr.str_to_pairlist(mail_mgr.get(filename, mailfile.FROM_L))
    self.reciplist = addr.str_to_pairlist(mail_mgr.get(filename, mailfile.TO_L))
    self.reciplist += addr.str_to_pairlist(mail_mgr.get(filename, mailfile.CC_L))
    self.reciplist += addr.str_to_pairlist(mail_mgr.get(filename, mailfile.BCC_L))

    self.tags = mail_mgr.get_tags(filename)

    temp = mail_mgr.get(filename, mailfile.BODY_L).split()
    self.words = [t.translate(punct_trans).lower() for t in temp]

    attach_s = mail_mgr.get(filename, mailfile.ATTACH_L).strip()
    self.num_attach = 0 if len(attach_s) == 0 else len(attach_s.split(", "))

    self.from_me = mail_mgr.get(filename, mailfile.SENT_L) == "True"

    self.datestr = mail_mgr.get(filename, mailfile.DATE_L)
    self.date_obj = None
    try:
      self.date_obj = dateutil.parser.parse(self.datestr)
    except:
      self.date_obj = datetime.datetime(year=9999,month=12,day=1)


def listname_to_getter(ln):
  if ln == "addresses":
    return lambda ms: ms.authlist + ms.reciplist
  elif ln == "authors":
    return lambda ms: ms.authlist
  elif ln == "recipients":
    return lambda ms: ms.reciplist
  elif ln == "tags":
    return lambda ms: ms.tags
  elif ln == "words":
    return lambda ms: ms.words
  else:  # shouldn't happen
    return lambda ms: [1]


def get_strfunc(ln):
  if ln in ["addresses", "authors", "recipients"]:
    return lambda pr: addr.pair_to_str(*pr)
  elif ln == "tags":
    return lambda tag: tag
  elif ln == "words":
    return lambda word: word
  else:  # shouldn't happen
    return lambda s: str(s)


# we need to count, sort, combine list of lists
# given [list1, list2, ...]
# return [(x1,n1), (x2,n2), ...]
# where x1 appears in a list n1 times, etc
def get_count_list(list_lists):
  counts = {}  # map x to number of occurrences
  for xs in list_lists:
    for x in xs:
      counts[x] = counts.get(x, 0) + 1
  return sorted(counts.items(), key = lambda pr: (-pr[1], pr[0]))


class MailStats:
  def __init__(self, mail_mgr):
    self.mail_mgr = mail_mgr
    self.msg_sums = {}       # map filename to summary

  # get message summary object
  def get_msg_sum(self, filename):
    if filename not in self.msg_sums:
      self.msg_sums[filename] = MsgSum(filename, self.mail_mgr)
    return self.msg_sums[filename]

  # get top entries by count
  def get_top(self, listname, filelist):
    if len(filelist) == 0:
      return []
    getter = listname_to_getter(listname)
    return get_count_list([getter(self.get_msg_sum(f)) for f in filelist])

  def get_stats(self, listname, filelist):
    getter = listname_to_getter(listname)
    mylists = [getter(self.get_msg_sum(f)) for f in filelist]
    counts = sorted([len(l) for l in mylists])
    return compute_stats(counts)
#  def get_stats(self, listname, filelist):
#    count_list = self.get_top(listname, filelist)
#    counts = [c[1] for c in count_list]
#    return compute_stats(counts)

  # return filenames sorted into each [bin_unit] in [slice_unit], over all slice_units
  # bin_unit can be hour, day, week, month, or year
  # slice_unit can be day, week, month, year, or None
  #
  # return list of bins sorted chronologically
  # each bin is a tuple (identifier, list of files, num_sliced_over)
  def get_time_slice(self, filenames, bin_unit, slice_unit=None):
    now = datetime.datetime.now()
    last_bin_obj = round_up(now, bin_unit)
    binstr_to_flist = {}
    for f in filenames:
      bound_str = round_up(self.get_msg_sum(f).date_obj, bin_unit).isoformat()[:13]
      if bound_str in binstr_to_flist:
        binstr_to_flist[bound_str].append(f)
      else:
        binstr_to_flist[bound_str] = [f]
    return do_slice(binstr_to_flist, bin_unit, slice_unit)

  # So the kinds of stats seem to be:
  # indicators (print total and fraction)
  # numbers, including length of lists (print general stats)
  # aggregations of lists, e.g. all authors (print top members and indicators)
  # lengths of *aggregations* of lists (print general stats)

  def get_summary(self, filenames):
    if len(filenames) == 0:
      return "No messages.\n"
    l = []
    l.append("\n")
    l.append("Number of messages: ")
    l.append(str(len(filenames)))
    l.append("\n\n")

    l.append("Popular tags: \n")
    l += itemlist_to_strlist(self.get_top("tags", filenames), lambda tag: tag)
    l.append("\n\n")

    for s in ["addresses", "authors", "recipients"]:
      l.append("All ")
      l.append(s)
      l.append(" by number of emails:\n")
      count_list = self.get_top(s, filenames)
      l += get_stats_strlist_from_counts([p[1] for p in count_list])
      l.append("Most common ")
      l.append(s)
      l.append(":\n")
      l += itemlist_to_strlist(count_list, lambda pr: addr.pair_to_str(*pr))
      l.append("\n\n")
    
    l.append("Word count:\n")
    l += get_stats_strlist(self.get_stats("words", filenames))
    l.append("\n")
    l.append("Popular words:\n")
    l += itemlist_to_strlist(self.get_top("words", filenames), lambda word: word)
    return "".join(l)


def parse_args(args):
  if len(args) == 0:
    return "default", None, None, None, None
  if args[0] == "slice":
    into_i = args.index("into")
    attr, listname, filt_args = parse_filt_args([] if into_i == 1 else args[1:into_i])
    bin_units = args[into_i+1]
    if not any([bin_units.startswith(s) for s in ["hour","day","week","month","year"]]):
      raise Exception("unrecognized bin unit in slice: " + bin_units)
    if len(args) > into_i+2:
      slice_units = args[-1]
      if not any([slice_units.startswith(s) for s in ["day","week","month","year"]]):
        raise Exception("unrecognized slice unit: " + slice_units)
    else:
      slice_units = None
  else:
    bin_units = None
    slice_units = None
    attr, listname, filt_args = parse_filt_args(args)
  return attr, listname, filt_args, bin_units, slice_units


def parse_filt_args(args):
  attr, listname, filt_args = "default", None, None
  if len(args) == 0:
    return attr, listname, filt_args

  if "matching" in args:
    i = args.index("matching")
    filt_args = args[i+1:]
    args = args[:i]
  if len(args) > 0:
    attr = args[0]
    if attr not in ["total", "fraction", "top", "num"]:
      raise Exception("Unrecognized attribute type: " + attr)
  if len(args) > 1 or attr == "top" or attr == "num":
    # if top or num, expect there to be another arg
    # (raise an exception if not)
    listname = args[1]
    if listname not in ["addresses","authors","recipients","tags","words"]:
      raise Exception("Unrecognized list: " + listname)
  return attr, listname, filt_args


# ms = MailStats object
def main(filelist, mail_mgr, ms, attr, listname, filt_args, bin_units, slice_units):
  match_list = filelist if filt_args is None else search.filter_arglist(filt_args, mail_mgr, filelist)
  if bin_units is None:
    if attr == "default":
      return ms.get_summary(match_list)   # matching ...
    elif attr == "total":
      return [str(len(match_list))]
    elif attr == "fraction":
      return [str(len(match_list) / len(filelist))]
    elif attr == "top":
      return itemlist_to_strlist(ms.get_top(listname, match_list), get_strfunc(listname))
    elif attr == "num":
      return get_stats_strlist(ms.get_stats(listname, match_list))
  else:  # SLICE
    binlist = ms.get_time_slice(match_list, bin_units, slice_units)
    labels = [b[0] for b in binlist]
    if attr == "default" or attr == "total":
      totals = [len(b[1]) for b in binlist]
      return get_bar_plot(labels, totals)
    elif attr == "fraction":
      all_binlist = ms.get_time_slice(filelist, bin_units, slice_units)
      fracs = [len(b[1]) / len(a[1]) for b,a in zip(binlist, all_binlist)]
      return get_bar_plot(labels, fracs)
    elif attr == "top":
      attrs = [ms.get_top(listname, b[1]) for b in binlist]
      return get_list_plot(labels, attrs, listname)
    elif attr == "num":
      stats = [ms.get_stats(listname, b[1]) for b in binlist]
      return get_mean_plot(labels, stats)
  return ["error\n"]


def process(args, filenames, mail_mgr, ms):
  try:
    attr, listname, filt_args, bin_units, slice_units = parse_args(args)
  except Exception as e:
    print("Error parsing: " + str(e))
    print(usage_str)
    return
  result_strlist = main(filenames, mail_mgr, ms, attr, listname, filt_args, bin_units, slice_units)
  print()
  print("".join(result_strlist))


def main_cli():
  global max_items, max_cols
  args = sys.argv[1:]
  if "-h" in args:
    print(usage_str)
    exit()

  if "-max_items" in args:
    ind = args.index("-max_items")
    if len(args) <= ind+1:
      print(usage_str)
      exit()
    max_items = max(int(args[ind+1]), 5)
    args = args[:ind] + ([] if len(args) <= ind+2 else args[ind+2:])

  if "-max_cols" in args:
    ind = args.index("-max_cols")
    if len(args) <= ind+1:
      print(usage_str)
      exit()
    max_cols = max(int(args[ind+1]), 5)
    args = args[:ind] + ([] if len(args) <= ind+2 else args[ind+2:])

  mail_mgr = mailfile.MailMgr()
  filenames = search.search_arglist(args, mail_mgr)
  ms = MailStats(mail_mgr)

  # interactive mode
  while True:
    command = input("\nType a command (or default, help, exit):\n")
    args = shlex.split(command)
    if len(args) > 0:
      if "-h" in args or args[0] == "help":
        print(help_str)
      elif args[0] == "exit" or args[0] == "quit":
        exit()
      elif args[0] == "default":
        process([], filenames, mail_mgr, ms)
      else:
        process(args, filenames, mail_mgr, ms)
 

if __name__ == "__main__":
  main_cli()





####
# bomail.util.datestuff
#
# Everything dealing with the mess that is date and time.
####

import datetime
import dateutil.parser
import pytz
from dateutil import tz
from dateutil.relativedelta import relativedelta
import traceback
import email, email.utils

datestr_str = """
'date strings' are used in search and scheduling, e.g.
    bomail search -a datestr   # search after given date
    bomail chstate -s datestr  # schedule for given date

datestr can be in:
 (1) Absolute format: yyyy-mm-ddTHH:MM
     or any prefix of this.

 (2) Relative format: p[num][unit]
     meaning plus num of units from now.
     Unit can be y (year), w (week), H (hour), M (minute).
     month is not allowed, to avoid confusing with minute.
     If num has a decimal point, adds the exact amount.
     Else, rounds down to the nearest unit.

 (3) Relative format:  m[num][unit]
     meaning minus num of units before now.
     Same details as (2) apply.

Some examples for absolute date:
  2050              # Jan 1st, 2050 at 00:00
  2050-03-07        # Mar 7th, 2050 at 00:00
  2050-03-07T11     # Mar 7th, 2050 at 11:00

Some examples for relative date:
  p1d               # tomorrow at 00:00, no matter what time it is today
  p1.d              # tomorrow at the same time of day that it is right now
  p1.5d             # 36 hours from now
  m3w               # 3 weeks before now, rounded down to Monday at 00:00
  m3.5w             # exactly 3.5 weeks before now (to the minute)
"""



# parse the email's date string and return a datetime object
def parse_email_date(s):
  if len(s) == 0:
    return datetime.datetime(9999,12,30)
  if "A.D." in s:  # remove this, is messes up the parser
    i = s.index("A.D.")
    s = s[:i] + (s[i+4:] if i+4 < len(s) else "")
  try:
    result = email.utils.parsedate_to_datetime(s)
    if result:  # if timezeone-unaware, should be UTC time
      if result.tzinfo is None:
        result = result.replace(tzinfo=pytz.utc)
      return result
    return datetime.datetime(9999,12,28)
  except:
    return datetime.datetime(9999,12,29)

# return datetime object for UTC time right now, timezone-aware
def get_utc_nowobj():
  return datetime.datetime.now(datetime.timezone.utc)

def parse_to_utc_obj(datestr):
  dateobj = dateutil.parser.parse(datestr)
  if dateobj.tzinfo is None:
    dateobj = dateobj.replace(tzinfo=pytz.utc)
  return dateobj.astimezone(pytz.utc)

def parse_to_utc_datestr(datestr):
  return parse_to_utc_obj(datestr).isoformat()

# convert object to local-time object
def parse_to_local_obj(datestr):
  dateobj = dateutil.parser.parse(datestr)
  if dateobj.tzinfo is None:
    dateobj = dateobj.replace(tzinfo=pytz.utc)
  return dateobj.astimezone(tz.tzlocal())

# convert object to local ISO string
def get_local_str(dateobj):
  loc = dateobj.astimezone(tz.tzlocal())
  return loc.isoformat()

def get_local_nowstr():
  return datetime.datetime.now(tz.tzlocal()).isoformat()

# filename is ...email/yyyy/mm-dd/x.email
# or .../yyyy/mm-dd/x.draft
def get_date_from_filename(filename):
  # filedir = os.path.dirname(filename)  # too slow
  slashind = filename.rfind("/")  # assume x does not contain a slash!!
  return filename[slashind-10:slashind].replace("/","-")


def datestr_matches(datestr, after, before):
  return ((after is None or datestr[:len(after)] >= after)
           and (before is None or datestr[:len(before)] <= before))


# ------------------------------------
# parsing schedstr strings


# given a 'schedstr' like yyyy-mm or m3.0d,
# turn it into an absolute date string yyyy-mm-ddT...
# in local time zone, throw an exception if unable
#def get_absstr_from_datestr(schedstr):
#  dateobj = get_datetime(schedstr)
#  if dateobj is None:
#    dateobj = datetime.datetime.now(tz.tzlocal())
#  try:
#    dateobj = dateobj.astimezone(tz.tzlocal())
#  except:
#    dateobj = pytz.utc.localize(dateobj)
#  return dateobj.isoformat()[:16]



def get_relativedelta_int(num, suffix):
  if suffix == "y":
    return relativedelta(years=int(num))
  elif suffix == "w":
    return relativedelta(weeks=int(num))
  elif suffix == "d":
    return relativedelta(days=int(num))
  elif suffix == "H":
    return relativedelta(hours=int(num))
  else:  # "M"
    return relativedelta(minutes=int(num))



def get_relativedelta_float(num, suffix):
  if num < 0:
    return -get_relativedelta_float(-num, suffix)
  
  inum = int(num)
  if suffix == "y":
    return relativedelta(years=inum) + get_relativedelta_float((num-inum)*365, "d")
  elif suffix == "w":
    return relativedelta(weeks=inum) + get_relativedelta_float((num-inum)*7, "d")
  elif suffix == "d":
    return relativedelta(days=inum) + get_relativedelta_float((num-inum)*24, "H")
  elif suffix == "H":
    return relativedelta(hours=inum) + get_relativedelta_float((num-inum)*60, "M")
  else:  # "M"
    return relativedelta(minutes=inum)

def get_relativedelta(num, suffix, is_float):
  if is_float:
    return get_relativedelta_float(num, suffix)
  else:
    return get_relativedelta_int(num, suffix)


# convert schedule string to datetime object in UTC
# or None if failure
def parse_schedstr_to_utcobj(schedstr):
  if schedstr[0] == "p":
    mult = 1.0  # forward in time
  elif schedstr[0] == "m":
    mult = -1.0  # backward in time
  else: # absolute date: parse it and return
    try:
      return parse_to_utc_obj(schedstr)
    except:
      return None

  # if we get to here, it's a relative
  nowobj = get_utc_nowobj()
  numstr = schedstr[1:-1]
  num = mult*float(numstr)  # forward or back in time
  suffix = schedstr[-1]
  diff = None
  is_float = "." in numstr
  result = nowobj + get_relativedelta(num, suffix, is_float)
  default_obj = datetime.datetime(1,1,1,tzinfo=pytz.utc)

  if not is_float:  # round the answer
    if suffix == "y":
      result = default_obj.replace(result.year)
    elif suffix == "w":
      dayoffset = relativedelta(days=result.weekday())
      result = default_obj.replace(year=result.year, month=result.month, day=result.day) - dayoffset
    elif suffix == "d":
      result = default_obj.replace(year=result.year, month=result.month, day=result.day)
    elif suffix == "H":
      result = default_obj.replace(year=result.year, month=result.month, day=result.day, hour=result.hour)
    elif suffix == "M":
      result = default_obj.replace(year=result.year, month=result.month, day=result.day, hour=result.hour, minute=result.minute)

  return result


def parse_schedstr_to_utcstr(schedstr):
  dateobj = parse_schedstr_to_utcobj(schedstr)
  return dateobj.isoformat()


# given date object, get date string to print in schedule file
def get_printed_schedulestr(dateobj):
  rounded_date = dateobj.replace(seconds=0, microseconds=0)
  return rounded_date.astimezone(pytz.utc).isoformat()







####
# bomail.util.listmail
#
# (Quickly) list emails matching date range or simple state queries.
# The idea is it's fast because it doesn't have to open the files.
####

import sys
import os.path

import bomail.config.config as config
import bomail.util.datestuff as datestuff
import bomail.cli.mailfile as mailfile

# after and before may be none
def list_all(after, before, mgr, dirname=config.email_dir):
  if not os.path.exists(dirname):
    return []
  all_years = os.listdir(dirname)
  if len(all_years) == 0:
    return []
  output = []
  for year_str in all_years:
    if after is not None and year_str < after[:4]:   continue
    if before is not None and year_str > before[:4]: continue
    year_prefix = year_str + "-"
    year_dir = dirname + year_str + "/"
    all_days = os.listdir(year_dir)
    for day_str in all_days:
      date_str = year_prefix + day_str
      if after is not None and date_str[:len(after)] < after:   continue
      if before is not None and date_str[:len(before)] > before: continue
      day_dir = year_dir + day_str + "/" 
      for s in os.listdir(day_dir):
        if len(s) >= 5 and (s[-4:] == "mail" or s[-5:] == "draft"):
          output.append(day_dir + s)
  return output

# openlist is sorted by reverse lexicographic order
def list_open(after, before, mgr):
  if not os.path.exists(config.openlist_file):
    return []
  output = []
  with open(config.openlist_file) as f:
    for line in f:
      filename = line[:-1] if (len(line) > 0 and line[-1] == "\n") else line
      if datestuff.datestr_matches(datestuff.get_date_from_filename(filename), after, before):
        output.append(filename)
  return output

# scheduledlist is sorted by schedule date, not file date!
# have to check date of file!
# lines have the form "scheduledate emailfilename"
def list_scheduled(after, before, mgr):
  if not os.path.exists(config.scheduledlist_file):
    return []
  output = []
  with open(config.scheduledlist_file) as f:
    for line in f:
      if len(line.strip()) == 0:
        continue
      filename = line.split()[1]
      if filename[-1] == "\n":
        filename = filename[:-1]
      ds = datestuff.get_date_from_filename(filename)
      if datestuff.datestr_matches(ds, after, before):
        output.append(filename)
  return output

def list_draft(after, before, mgr):
  return list_all(after, before, mgr, config.drafts_dir)

# similar to list_all, but use the attach directories (only the ones that exist)
def list_attachonly(after, before, mgr):
  if not os.path.exists(config.attach_dir):
    return []
  all_years = os.listdir(config.attach_dir)
  if len(all_years) == 0:
    return []
  output = []
  for year_str in all_years:
    if after is not None and year_str < after[:4]:   continue
    if before is not None and year_str > before[:4]: continue
    year_prefix = year_str + "-"
    year_dir = config.attach_dir + year_str + "/"
    all_days = os.listdir(year_dir)
    for day_str in all_days:
      date_str = year_prefix + day_str
      if after is not None and date_str[:len(after)] < after:   continue
      if before is not None and date_str[:len(before)] > before: continue
      day_dir = year_dir + day_str + "/" 
      for s in os.listdir(day_dir):
        if s[-1] == "/":
          s = s[:-1]
        mailfile_path = config.email_dir + year_str + "/" + day_str + "/" + s + ".mail"
        output.append(mailfile_path)
  return output
   


class ListOpts:
  def __init__(self):
    self.after_str = None
    self.before_str = None
    self.source = "all"

  def listmail(self, mgr):
    if self.source == "all":
      return list_all(self.after_str, self.before_str, mgr)
    elif self.source == "open":
      return list_open(self.after_str, self.before_str, mgr)
    elif self.source == "scheduled":
      return list_scheduled(self.after_str, self.before_str, mgr)
    elif self.source == "draft":
      return list_draft(self.after_str, self.before_str, mgr)
    elif self.source == "attach":
      return list_attachonly(self.after_str, self.before_str, mgr)





####
# bomail.util.attach
#
# Utilities for dealing with attachments.
####

# given attachment paths, convert to a string to be stored in a file
# by quoting each and escaping inner quotes, then comma-and-space separating
def attach_paths_to_str(pathlist):
  anslist = ['"' + path.replace('\\','\\\\').replace('"','\\"') + '"' for path in pathlist]
  return ', '.join(anslist)

# given a string formatted as above,
# parse into a list of paths
def attach_str_to_paths(s, initial_pathlist = None):
  pathlist = [] if initial_pathlist is None else initial_pathlist
  if len(s.strip()) == 0:
    return pathlist
  try:
    # find first "
    i = s.index('"')
    # find matching " by skipping anything escaped
    j = i+1
    while s[j] != '"':
      if s[j] == '\\':
        j += 1
      j += 1
    mys = s[i+1:j].replace('\\"','"').replace('\\\\','\\')
    pathlist.append(mys)
    attach_str_to_paths(s[j+1:], pathlist)
    return pathlist
  except:
    pathlist.append(s)
    return pathlist




####
# bomail.cli.clean_tags
#
# A little utility to read through all emails, find any tags
# that are unused, and remove them from the tag list.
####

import os,sys

import bomail.cli.mailfile as mailfile
import bomail.cli.search as search
from bomail.util.tags import TagMgr

usage_str = """Scans all email for saved tags that aren't being used,
and deletes them."""

# maximum messages to hold in memory before releasing them
MAX_MSGS_IN_MEM = 10000


def main_cli():
  if "-h" in sys.argv:
    print(usage_str)
    exit()

  mgr = mailfile.MailMgr()
  all_fnames = search.search_argstr("", mgr)
  used_tagset = set()
  msgs_in_mem = 0
  for fname in all_fnames:
    my_tagset = mgr.get_tagset(fname)
    if "" in my_tagset:
      print(fname)
    used_tagset.update(my_tagset)
    msgs_in_mem += 1
    if msgs_in_mem >= MAX_MSGS_IN_MEM:
      mgr = mailfile.MailMgr()  # release memory of older messages

  tagmgr = TagMgr()
  tagmgr.reset_tags_to(list(used_tagset))


if __name__ == "__main__":
  main_cli()





import os, sys

import bomail.config.config as config
from bomail.cli import gui,process,check_sched,search,chstate,mailfile,compose,clean_tags,send,stats
import bomail.util.datestuff as datestuff

usage_str = """
Usage:
Use bomail <command> -h for help on that command.

Commands:
  bomail gui         Launch interface (if in doubt, run this)

  bomail process     Check for and process new emails
  bomail check_sched Check 'scheduled' emails and unschedule them
  bomail search      Search
  bomail chstate     Change state of emails (open/closed/scheduled/trash)
  bomail compose     Compose new emails as blank, reply, forward
  bomail mailfile    See fields of an email or add/remove tags
  bomail send        Send drafts
  bomail stats       See some stats about email

  bomail help datestr  Print help about using relative dates
  bomail help tags     Print help about 'dirlike' tags

Configuration file: ~/.bomailrc
Data location: """ + config.bomail_data_base + """

"""

extra_info = """
Note many commands read a list of filenames from stdin,
so for example to tag a set of emails 'food', you can use:
  bomail search chocolate | bomail mailfile -add-tags food

"""

taghelp_str = """
Using tags:

An email can have unlimited tags.
Searching for emails with a given tag must match
exactly (so "bomail search -t happy" does not match
emails tagged happydays, but does match emails tagged happy.)

Twist: 'directory-like' tags.
Whenever your tag has a / character, it matches searches
for tags of a prefix up to a /.

Example: anything tagged social/twitter, social/facebook,
or social/mastodon will match "bomail search -t social".
However, none of those will match "bomail search -t facebook".
(Use "bomail search -t social/facebook" instead.)

"""


def main():
  if len(sys.argv) <= 1:
    print(usage_str)
    exit(0)

  if sys.argv[1] == "help":
    if len(sys.argv) <= 2:
      print(usage_str)
    elif sys.argv[2] == "datestr":
      print(bomail.util.datestuff.datestr_str)
    elif sys.argv[2] == "tags":
      print(taghelp_str)
    else:
      print(usage_str)
    exit(0)

  # hacky but simple: make programs act like they were
  # invoked from command line
  sys.argv = sys.argv[1:]

  if sys.argv[0] == "gui":
    gui.main_cli()
  elif sys.argv[0] == "process":
    process.main_cli()
  elif sys.argv[0] == "check_sched":
    check_sched.main_cli()
  elif sys.argv[0] == "search":
    search.main_cli()
  elif sys.argv[0] == "chstate":
    chstate.main_cli()
  elif sys.argv[0] == "mailfile":
    mailfile.main_cli()
  elif sys.argv[0] == "compose":
    compose.main_cli()
  elif sys.argv[0] == "clean_tags":
    clean_tags.main_cli()
  elif sys.argv[0] == "send":
    send.main_cli()
  elif sys.argv[0] == "stats":
    stats.main_cli()

  else:
    print(usage_str)

if __name__ == "__main__":
  main()




####
# bomail.config.config
#
# Main configuration file.
# When loaded, it tries to open ~/.bomailrc and read it; if unsuccessful,
# it runs a short setup program asking questions.
####

import os, sys

bomailrc_filename = ".bomailrc"


#--------------------------------------------------------------
setup_success_str = """
Basic setup successful!
 (1) Edit ~/.bomailrc in order to send email and other options.
 (2) Edit ~/bomail/metadata/mail-handlers.txt to set up auto-tagging
     and processing of new mail.

Run bomail to begin...
"""


#--------------------------------------------------------------
bomail_tabcomplete_str = """

_bomail()
{
    local cur=${COMP_WORDS[COMP_CWORD]}
    COMPREPLY=( $(compgen -W "gui process check_sched search chstate compose mailfile send stats clean_tags" -- $cur) )
}
complete -F _bomail bomail

"""


#--------------------------------------------------------------
sample_mailhandlers_str = """## This file controls handling of new emails in bomail.
## In this file, lines starting with '#' are comments.

## ------------------------------------------------------
## Section 1: Special options
##
## Turn each off by commenting it with '#'.
## ------------------------------------------------------

## Tag replies to a thread with parent message's tags
-autotagreplies

## Use tags from last line of email if of the form "tags: tag1, ..."
-lastlinetags

## Automatically close any sent mail
-close-sent


## ---------------------------------------------------
## Section 2: Custom handlers
##
## You write these.
## ---------------------------
##
## Each handler is a list of search strings (see bomail search -h)
## followed by a list of actions to take for those messages.
## Actions are "open", "close", "schedule schedstr" (see bomail help datestr),
## and "tag tag1, tag2, ...". Examples:
##
## -from Facebook
## tag social
##
## -from Instagram
## OR -subject Instagram
## close
## AND tag social
"""

#--------------------------------------------------------------
sample_config_str = """## bomail configuration file
##
## All lines beginning with # are comments and are ignored.
## To uncomment a line, delete the initial #.



## ============================================================
## User info: must edit this section!

name = YOUR NAME
email_addr = USER@DOMAIN.COM

## Send email from this server.
## If smtp_servername = localhost, the other options are ignored.
## For SMTP, read username/pass from file, or enter them here.
smtp_servername = smtp.gmail.com
smtp_port = 587
smtp_userpass_file = /home/LINUXUSERNAME/.getmail/getmailrc
#smtp_username = USER or USER@DOMAIN.COM
#smtp_password = PASSWORD

#alias_addresses = ALSO_ME@DOMAIN.COM, ALSO_METOO@DOMAIN.COM

## Additional aliases here. Add numbers 2,3,...
## Just edit "From" line in email to send from these instead.
#name2 = YOUR NAME
#email_addr2 = ALTERNATE@DOMAIN.COM
#smtp_servername2 = smtp.office365.com
#smtp_port2 = 587
#smtp_userpass_file2 = /home/bo/.getmail/getmailrc_bowaggonergmail
#smtp_username2 = USER2 or USER2@DOMAIN.COM
#smtp_password2 = PASSWORD2



## ============================================================
## Organization options: should confirm these are okay

bomail_location = ~/bomail
new_rawmail_location = ~/mail/new
processed_rawmail_location = ~/mail/cur



## ============================================================
## Important UI options

## Comment next line to turn off threading (conversations)
threads_on

read_program = less
edit_program = vim

## Comment next line to use vim-style navigation (hjkl for left/down/up/right)
#arrowkey_navigation

## Options are light1, light2, dark1, dark2
colorscheme = dark1



## ============================================================
## Other UI options

## Options are none, some, all
strip_newlines_in_msg_preview = some

## Num lines (including subject/author/date and preview lines) per message
total_lines_per_msg = 5
threadview_total_lines_per_msg = 10

## Number of characters to use ("len") and for spacing ("pad")
date_len = 20
date_pad = 2
author_len = 24
author_pad = 2

## Comment to turn off
#hline_between_msgs
skipline_between_msgs
show_tags
tags_on_topline


"""
#--------------------------------------------------------------


home = os.getenv("HOME")
config_file = os.path.join(home, bomailrc_filename)


# Initial setup!
if not os.path.exists(config_file):
  print("\nWelcome to bomail!")

  sys.stdout.write("Name (as it will appear in From line): ")
  sys.stdout.flush()
  myname = sys.stdin.readline().strip()
  sys.stdout.write("Email address (primary, i.e. send email from): ")
  sys.stdout.flush()
  myaddr = sys.stdin.readline().strip()

  default_datadir = os.path.join(home,"bomail")
  sys.stdout.write("Data files location? (leave blank to use " + default_datadir +"): ")
  sys.stdout.flush()
  datadir = sys.stdin.readline().strip()
  if len(datadir) == 0:
    datadir = default_datadir

  
  default_newmail_dir = os.path.join(home, "mail/new")
  sys.stdout.write("Process new mail from where? (leave blank to use " + default_newmail_dir + "): ")
  sys.stdout.flush()
  newmail_dir = sys.stdin.readline().strip()
  if len(newmail_dir) == 0:
    newmail_dir = default_newmail_dir

  default_oldmail_dir = os.path.join(home, "mail/cur")
  sys.stdout.write("Put mail where after processing? (leave blank to use " + default_oldmail_dir + "): ")
  sys.stdout.flush()
  oldmail_dir = sys.stdin.readline().strip()
  if len(oldmail_dir) == 0:
    oldmail_dir = default_oldmail_dir

  sys.stdout.write("Add bomail command tab-completion to your ~/.bashrc file? (y/n): ")
  sys.stdout.flush()
  tabcomplete = sys.stdin.readline().strip()
  if len(tabcomplete) > 0 and tabcomplete[0] in ["y", "Y"]:
    with open(os.path.join(home, ".bashrc"), "a") as f:
      f.write(bomail_tabcomplete_str)

  # this is not ideal because it doesn't change if the configuration at
  # the bottom of the file changes
  metadata_dir  = os.path.join(datadir, "metadata/")
  handlers_file = os.path.join(metadata_dir, "mail-handlers.txt")
  if not os.path.exists(handlers_file):
    os.makedirs(metadata_dir, exist_ok=True)
    with open(handlers_file, "w") as f:
      f.write(sample_mailhandlers_str)

  with open(config_file, "w") as f:
    f.write(sample_config_str.replace("YOUR NAME", myname).replace("USER@DOMAIN.COM", myaddr).replace("~/bomail", datadir).replace("~/mail/new",newmail_dir).replace("~/mail/cur",oldmail_dir))

  print(setup_success_str)
  exit(0)


# Function to open and parse config file
def parse_config_file():
  options_dict = {}
  def parse_line(line):
    line = line.strip()
    if len(line) > 0 and line[0] != "#":
      if "=" in line:
        left, right = line.split("=")
        options_dict[left.strip()] = right.strip()
      else:
        options_dict[line.strip()] = True
  
  with open(config_file) as f:
    for line in f:
      parse_line(line)
  return options_dict


# Actually do it
options_dict = parse_config_file()


# --------------------------------------------------------------
# User info

names = []
email_addrs = []
smtp_servernames = []
smtp_ports = []
smtp_usernames = []
smtp_passwords = []

# return True on success, False on failure
def get_user_info(j):
  suffix = "" if j==0 else str(j)
  if not ("name" + suffix) in options_dict:
    return False
  names.append(options_dict["name" + suffix])
  email_addrs.append(options_dict["email_addr" + suffix])
  smtp_servernames.append(options_dict["smtp_servername" + suffix])
  if smtp_servernames[-1] == "localhost":
    return True
  smtp_ports.append(int(options_dict["smtp_port" + suffix]))
  if ("smtp_userpass_file" + suffix) in options_dict:
    try:
      with open(options_dict["smtp_userpass_file" + suffix]) as f:
        for line in f:
          if line.startswith("username ="):
            uname = line[line.index("=")+1:].strip()
          if line.startswith("password ="):
            pword = line[line.index("=")+1:].strip()
      smtp_usernames.append(uname)
      smtp_passwords.append(pword)
    except:
      raise Exception("Could not read username/password from " + options_dict["smtp_userpass_file" + suffix])
  else: # get username and password straight from the config
    smtp_usernames.append(options_dict["smtp_username" + suffix])
    smtp_passwords.append(options_dict["smtp_password" + suffix])
  return True

# first get empty suffix, then 2, 3, ...
get_user_info(0)
j = 2
while get_user_info(j):
  j += 1

my_aliases = []
if "alias_addresses" in options_dict:
  my_aliases = [s.strip() for s in options_dict["alias_addresses"].split(",")]



# --------------------------------------------------------------
# Organization locations

def add_slash(s):
  return s if s[-1] == "/" else s + "/"

bomail_data_base = add_slash(options_dict["bomail_location"].replace("~", home))
new_rawmail_dir = add_slash(options_dict["new_rawmail_location"].replace("~", home))
old_rawmail_dir = add_slash(options_dict["processed_rawmail_location"].replace("~", home))

# save a copy of sent messages here if sending from localhost
save_sent_dir = new_rawmail_dir


# --------------------------------------------------------------
# Options not in bomailrc (may edit)

# how many new emails to process at a time, -1 for no limit
process_new_limit = -1

# how many new emails to process in a batch, decrease to save memory
# for very large imports
process_batch_size = 1000

# tag drafts with same tags as parent
autotag_draft_replies = True

# Replying to emails
# will replace %from, %to, %date, %time, %body
# with the corresponding text from the quoted email
quote_format = """\n
%from wrote on %date at %time:
%body"""

# %a is day of week; %Y,%m,%d are year/month/date
# %H,%M are hour/minute
quote_date_fmt = "%a, %Y-%m-%d"  
quote_time_fmt = "%H:%M"         
quote_line_prefix = "> "


# --------------------------------------------------------------
# File locations: suggested not to edit

# Locations of directories
email_dir     = os.path.join(bomail_data_base, "email/")
attach_dir    = os.path.join(bomail_data_base, "attachments/")
drafts_dir    = os.path.join(bomail_data_base, "drafts/")
trash_dir     = os.path.join(bomail_data_base, "trash/")
metadata_dir  = os.path.join(bomail_data_base, "metadata/")

# Locations of files that must be edited by hand
handlers_file      = os.path.join(metadata_dir, "mail-handlers.txt")
addr_alias_file    = os.path.join(metadata_dir, "aliases.txt")

# Locations of files that may be edited by hand
tags_file          = os.path.join(metadata_dir, "tags.txt")
addr_book_file     = os.path.join(metadata_dir, "addr_book.txt")

# Locations of files that should not be hand-edited
openlist_file      = os.path.join(metadata_dir, "openlist.txt")
scheduledlist_file = os.path.join(metadata_dir, "scheduledlist.txt")
msg_ids_file       = os.path.join(metadata_dir, "msg_ids.txt")
tab_config_file    = os.path.join(metadata_dir, "tab_config.py")

# Locations of logs, these can be deleted whenever
acts_log_file      = os.path.join(metadata_dir, "acts_log.txt")
error_log_file     = os.path.join(metadata_dir, "error_log.txt")

for d in [email_dir, attach_dir, drafts_dir, trash_dir, metadata_dir]:
  if not os.path.exists(d):
    os.makedirs(d)

